# -*- coding: utf-8 -*-

# This code is part of Ansible, but is an independent component.
# This particular file snippet, and this file snippet only, is BSD licensed.
# Modules you write using this snippet, which is embedded dynamically by Ansible
# still belong to the author of the module, and may assign their own license
# to the complete work.
#
# Copyright (c), Michael DeHaan <michael.dehaan@gmail.com>, 2012-2013
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""
This module adds helper functions for various EC2 specific services.

It also includes a large number of imports for functions which historically
lived here.  Most of these functions were not specific to EC2, they ended
up in this module because "that's where the AWS code was" (originally).
"""

import copy
import re
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import ansible.module_utils.common.warnings as ansible_warnings  # pylint: disable=unused-import
from ansible.module_utils.ansible_release import __version__

# Used to live here, moved into ansible.module_utils.common.dict_transformations
from ansible.module_utils.common.dict_transformations import _camel_to_snake  # pylint: disable=unused-import
from ansible.module_utils.common.dict_transformations import _snake_to_camel  # pylint: disable=unused-import
from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict  # pylint: disable=unused-import
from ansible.module_utils.common.dict_transformations import snake_dict_to_camel_dict  # pylint: disable=unused-import

# Used to live here, moved into ansible_collections.amazon.aws.plugins.module_utils.arn
from .arn import is_outpost_arn as is_outposts_arn  # pylint: disable=unused-import
from .arn import validate_aws_arn

# Used to live here, moved into ansible_collections.amazon.aws.plugins.module_utils.botocore
from .botocore import HAS_BOTO3  # pylint: disable=unused-import
from .botocore import boto3_conn  # pylint: disable=unused-import
from .botocore import boto3_inventory_conn  # pylint: disable=unused-import
from .botocore import boto_exception  # pylint: disable=unused-import
from .botocore import get_aws_connection_info  # pylint: disable=unused-import
from .botocore import get_aws_region  # pylint: disable=unused-import
from .botocore import is_boto3_error_code
from .botocore import paginated_query_with_retries
from .errors import AWSErrorHandler

# Used to live here, moved into ansible_collections.amazon.aws.plugins.module_utils.exceptions
from .exceptions import AnsibleAWSError  # pylint: disable=unused-import
from .iam import list_iam_instance_profiles

# Used to live here, moved into ansible_collections.amazon.aws.plugins.module_utils.modules
# The names have been changed in .modules to better reflect their applicability.
from .modules import _aws_common_argument_spec as aws_common_argument_spec  # pylint: disable=unused-import
from .modules import aws_argument_spec as ec2_argument_spec  # pylint: disable=unused-import

# Used to live here, moved into ansible_collections.amazon.aws.plugins.module_utils.policy
from .policy import _py3cmp as py3cmp  # pylint: disable=unused-import
from .policy import compare_policies  # pylint: disable=unused-import

# Used to live here, moved into ansible_collections.amazon.aws.plugins.module_utils.retries
from .retries import AWSRetry  # pylint: disable=unused-import

# Used to live here, moved into ansible_collections.amazon.aws.plugins.module_utils.tagging
from .tagging import ansible_dict_to_boto3_tag_list  # pylint: disable=unused-import
from .tagging import boto3_tag_list_to_ansible_dict  # pylint: disable=unused-import
from .tagging import boto3_tag_specifications
from .tagging import compare_aws_tags  # pylint: disable=unused-import

# Used to live here, moved into ansible_collections.amazon.aws.plugins.module_utils.transformation
from .transformation import ansible_dict_to_boto3_filter_list  # pylint: disable=unused-import
from .transformation import map_complex_type  # pylint: disable=unused-import

try:
    import botocore
except ImportError:
    pass  # Handled by HAS_BOTO3


class AnsibleEC2Error(AnsibleAWSError):
    pass


EC2TagSpecifications = Dict[str, Union[str, List[Dict[str, str]]]]


@AWSRetry.jittered_backoff()
def describe_availability_zones(
    client, **params: Dict[str, Union[List[str], bool, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Union[str, List[Dict[str, str]]]]]:
    # The paginator does not exist for `describe_availability_zones()`
    return client.describe_availability_zones(**params)["AvailabilityZones"]


@AWSRetry.jittered_backoff()
def describe_regions(
    client, **params: Dict[str, Union[List[str], bool, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, str]]:
    # The paginator does not exist for `describe_regions()`
    return client.describe_regions(**params)["Regions"]


# EC2 VPC Subnets
class EC2VpcSubnetErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidSubnetID.NotFound")


@EC2VpcSubnetErrorHandler.deletion_error_handler("delete subnet")
@AWSRetry.jittered_backoff()
def delete_subnet(client, subnet_id: str) -> bool:
    client.delete_subnet(SubnetId=subnet_id)
    return True


@EC2VpcSubnetErrorHandler.list_error_handler("describe subnets", [])
@AWSRetry.jittered_backoff()
def describe_subnets(
    client, **params: Dict[str, Union[List[str], bool, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_subnets")
    return paginator.paginate(**params).build_full_result()["Subnets"]


@EC2VpcSubnetErrorHandler.common_error_handler("create subnet")
@AWSRetry.jittered_backoff()
def create_subnet(client, **params: Dict[str, Union[str, bool, int, EC2TagSpecifications]]) -> Dict[str, Any]:
    return client.create_subnet(**params)["Subnet"]


@EC2VpcSubnetErrorHandler.common_error_handler("modify subnet")
@AWSRetry.jittered_backoff()
def modify_subnet_attribute(client, subnet_id: str, **params: Dict[str, Union[str, int, Dict[str, bool]]]) -> bool:
    client.modify_subnet_attribute(SubnetId=subnet_id, **params)
    return True


@EC2VpcSubnetErrorHandler.common_error_handler("disassociate subnet cidr block")
@AWSRetry.jittered_backoff()
def disassociate_subnet_cidr_block(client, association_id: str) -> Dict[str, Union[str, Dict[str, str]]]:
    return client.disassociate_subnet_cidr_block(AssociationId=association_id)["Ipv6CidrBlockAssociation"]


@EC2VpcSubnetErrorHandler.common_error_handler("associate subnet cidr block")
@AWSRetry.jittered_backoff()
def associate_subnet_cidr_block(
    client, subnet_id: str, **params: Dict[str, Union[str, int]]
) -> Dict[str, Union[str, Dict[str, str]]]:
    return client.associate_subnet_cidr_block(SubnetId=subnet_id, **params)["Ipv6CidrBlockAssociation"]


# EC2 VPC Route table
class EC2VpcRouteTableErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidRouteTableID.NotFound")


@EC2VpcRouteTableErrorHandler.list_error_handler("describe route tables", [])
@AWSRetry.jittered_backoff()
def describe_route_tables(
    client, **params: Dict[str, Union[List[str], bool, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_route_tables")
    return paginator.paginate(**params).build_full_result()["RouteTables"]


@EC2VpcRouteTableErrorHandler.common_error_handler("disassociate route table")
@AWSRetry.jittered_backoff()
def disassociate_route_table(client, association_id: str) -> bool:
    client.disassociate_route_table(AssociationId=association_id)
    return True


@EC2VpcRouteTableErrorHandler.common_error_handler("associate route table")
@AWSRetry.jittered_backoff()
def associate_route_table(
    client, route_table_id: str, **params: Dict[str, str]
) -> Dict[str, Union[str, Dict[str, str]]]:
    return client.associate_route_table(RouteTableId=route_table_id, **params)


@EC2VpcRouteTableErrorHandler.common_error_handler("enable vgw route propagation")
@AWSRetry.jittered_backoff()
def enable_vgw_route_propagation(client, gateway_id: str, route_table_id: str) -> bool:
    client.enable_vgw_route_propagation(RouteTableId=route_table_id, GatewayId=gateway_id)
    return True


@EC2VpcRouteTableErrorHandler.deletion_error_handler("delete route table")
@AWSRetry.jittered_backoff()
def delete_route_table(client, route_table_id: str) -> bool:
    client.delete_route_table(RouteTableId=route_table_id)
    return True


@EC2VpcRouteTableErrorHandler.common_error_handler("create route table")
@AWSRetry.jittered_backoff()
def create_route_table(client, vpc_id: str, tags: Optional[Dict[str, str]]) -> Dict[str, Any]:
    params = {"VpcId": vpc_id}
    if tags:
        params["TagSpecifications"] = boto3_tag_specifications(tags, types="route-table")
    return client.create_route_table(**params)["RouteTable"]


# EC2 VPC Route table Route
class EC2VpcRouteTableRouteErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidRoute.NotFound")


@EC2VpcRouteTableRouteErrorHandler.deletion_error_handler("delete route")
@AWSRetry.jittered_backoff()
def delete_route(client, route_table_id: str, **params: Dict[str, str]) -> bool:
    client.delete_route(RouteTableId=route_table_id, **params)
    return True


@EC2VpcRouteTableRouteErrorHandler.common_error_handler("replace route")
@AWSRetry.jittered_backoff()
def replace_route(client, route_table_id: str, **params: Dict[str, Union[str, bool]]) -> bool:
    client.replace_route(RouteTableId=route_table_id, **params)
    return True


@EC2VpcRouteTableRouteErrorHandler.common_error_handler("create route")
@AWSRetry.jittered_backoff()
def create_route(client, route_table_id: str, **params: Dict[str, str]) -> bool:
    return client.create_route(RouteTableId=route_table_id, **params)["Return"]


# EC2 VPC
class EC2VpcErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidVpcID.NotFound")


@EC2VpcErrorHandler.list_error_handler("describe vpcs", [])
@AWSRetry.jittered_backoff()
def describe_vpcs(
    client, **params: Dict[str, Union[List[str], bool, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_vpcs")
    return paginator.paginate(**params).build_full_result()["Vpcs"]


@EC2VpcErrorHandler.deletion_error_handler("delete vpc")
@AWSRetry.jittered_backoff()
def delete_vpc(client, vpc_id: str) -> bool:
    client.delete_vpc(VpcId=vpc_id)
    return True


@EC2VpcErrorHandler.common_error_handler("describe vpc attribute")
@AWSRetry.jittered_backoff()
def describe_vpc_attribute(client, vpc_id: str, attribute: str) -> Dict[str, Any]:
    # The paginator does not exist for `describe_vpc_attribute`
    return client.describe_vpc_attribute(VpcId=vpc_id, Attribute=attribute)


@EC2VpcErrorHandler.common_error_handler("modify vpc attribute")
@AWSRetry.jittered_backoff()
def modify_vpc_attribute(client, vpc_id: str, **params: Dict[str, Union[str, Dict[str, bool]]]) -> bool:
    client.modify_vpc_attribute(VpcId=vpc_id, **params)
    return True


@EC2VpcErrorHandler.common_error_handler("create vpc")
@AWSRetry.jittered_backoff()
def create_vpc(client, **params: Dict[str, Union[str, bool, int, EC2TagSpecifications]]) -> Dict[str, Any]:
    return client.create_vpc(**params)["Vpc"]


@EC2VpcErrorHandler.common_error_handler("associate vpc cidr block")
@AWSRetry.jittered_backoff()
def associate_vpc_cidr_block(client, vpc_id: str, **params: Dict[str, Union[str, bool, int]]) -> Dict[str, Any]:
    return client.associate_vpc_cidr_block(VpcId=vpc_id, **params)


@EC2VpcErrorHandler.common_error_handler("disassociate vpc cidr block")
@AWSRetry.jittered_backoff()
def disassociate_vpc_cidr_block(client, association_id: str) -> Dict[str, Any]:
    return client.disassociate_vpc_cidr_block(AssociationId=association_id)


# EC2 VPC Peering Connection
class EC2VpcPeeringErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidVpcPeeringConnectionID.NotFound")


@EC2VpcPeeringErrorHandler.list_error_handler("describe vpc peering", [])
@AWSRetry.jittered_backoff()
def describe_vpc_peering_connections(client, **params: Dict[str, Any]) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_vpc_peering_connections")
    return paginator.paginate(**params).build_full_result()["VpcPeeringConnections"]


@EC2VpcSubnetErrorHandler.common_error_handler("create vpc peering")
@AWSRetry.jittered_backoff()
def create_vpc_peering_connection(
    client, **params: Dict[str, Union[str, bool, int, EC2TagSpecifications]]
) -> Dict[str, Any]:
    return client.create_vpc_peering_connection(**params)["VpcPeeringConnection"]


@EC2VpcSubnetErrorHandler.deletion_error_handler("delete vpc peering")
@AWSRetry.jittered_backoff()
def delete_vpc_peering_connection(client, peering_id: str) -> bool:
    client.delete_vpc_peering_connection(VpcPeeringConnectionId=peering_id)
    return True


@EC2VpcSubnetErrorHandler.deletion_error_handler("accept vpc peering")
@AWSRetry.jittered_backoff()
def accept_vpc_peering_connection(client, peering_id: str) -> bool:
    client.accept_vpc_peering_connection(VpcPeeringConnectionId=peering_id)
    return True


@EC2VpcSubnetErrorHandler.deletion_error_handler("reject vpc peering")
@AWSRetry.jittered_backoff()
def reject_vpc_peering_connection(client, peering_id: str) -> bool:
    client.reject_vpc_peering_connection(VpcPeeringConnectionId=peering_id)
    return True


# EC2 vpn
class EC2VpnErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code(["InvalidVpnConnectionID.NotFound", "InvalidRoute.NotFound"])


@EC2VpcErrorHandler.list_error_handler("describe vpn connections", [])
@AWSRetry.jittered_backoff()
def describe_vpn_connections(client, **params: Dict[str, Any]) -> List[Dict[str, Any]]:
    # The paginator does not exist for `describe_vpn_connections`
    return client.describe_vpn_connections(**params)["VpnConnections"]


@EC2VpcErrorHandler.common_error_handler("create vpn connection route")
@AWSRetry.jittered_backoff()
def create_vpn_connection_route(client, vpn_connection_id: str, route: Dict[str, Any]) -> bool:
    client.create_vpn_connection_route(VpnConnectionId=vpn_connection_id, DestinationCidrBlock=route)
    return True


@EC2VpcErrorHandler.deletion_error_handler("delete vpn connection route")
@AWSRetry.jittered_backoff()
def delete_vpn_connection_route(client, vpn_connection_id: str, route: Dict[str, Any]) -> bool:
    client.delete_vpn_connection_route(VpnConnectionId=vpn_connection_id, DestinationCidrBlock=route)
    return True


@EC2VpcErrorHandler.common_error_handler("create vpn connection")
@AWSRetry.jittered_backoff()
def create_vpn_connection(client, **params: Dict[str, Any]) -> Dict[str, Any]:
    return client.create_vpn_connection(**params)["VpnConnection"]


@EC2VpcErrorHandler.deletion_error_handler("delete vpn connection")
@AWSRetry.jittered_backoff()
def delete_vpn_connection(client, vpn_connection_id: str) -> Dict[str, Any]:
    client.delete_vpn_connection(VpnConnectionId=vpn_connection_id)
    return True


# EC2 Internet Gateway
class EC2InternetGatewayErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidInternetGatewayID.NotFound")


@EC2InternetGatewayErrorHandler.list_error_handler("describe internet gateways")
@AWSRetry.jittered_backoff()
def describe_internet_gateways(
    client, **params: Dict[str, Union[List[str], bool, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_internet_gateways")
    return paginator.paginate(**params).build_full_result()["InternetGateways"]


@EC2InternetGatewayErrorHandler.common_error_handler("create internet gateway")
@AWSRetry.jittered_backoff()
def create_internet_gateway(client, tags: Optional[List[Dict[str, str]]]) -> Dict[str, Any]:
    params = {}
    if tags:
        params["TagSpecifications"] = boto3_tag_specifications(tags, types="internet-gateway")
    return client.create_internet_gateway(**params)["InternetGateway"]


@EC2InternetGatewayErrorHandler.common_error_handler("detach internet gateway")
@AWSRetry.jittered_backoff()
def detach_internet_gateway(client, internet_gateway_id: str, vpc_id: str) -> bool:
    client.detach_internet_gateway(InternetGatewayId=internet_gateway_id, VpcId=vpc_id)
    return True


@EC2InternetGatewayErrorHandler.common_error_handler("attach internet gateway")
@AWSRetry.jittered_backoff()
def attach_internet_gateway(client, internet_gateway_id: str, vpc_id: str) -> bool:
    client.attach_internet_gateway(InternetGatewayId=internet_gateway_id, VpcId=vpc_id)
    return True


@EC2InternetGatewayErrorHandler.deletion_error_handler("delete internet gateway")
@AWSRetry.jittered_backoff()
def delete_internet_gateway(client, internet_gateway_id: str) -> bool:
    client.delete_internet_gateway(InternetGatewayId=internet_gateway_id)
    return True


# EC2 NAT Gateway
class EC2NatGatewayErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidNatGatewayID.NotFound")


@EC2NatGatewayErrorHandler.list_error_handler("describe nat gateways", [])
@AWSRetry.jittered_backoff()
def describe_nat_gateways(
    client, **params: Dict[str, Union[List[str], bool, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_nat_gateways")
    return paginator.paginate(**params).build_full_result()["NatGateways"]


@EC2NatGatewayErrorHandler.deletion_error_handler("delete nat gateway")
@AWSRetry.jittered_backoff()
def delete_nat_gateway(client, nat_gateway_id: str) -> bool:
    client.delete_nat_gateway(NatGatewayId=nat_gateway_id)
    return True


@EC2NatGatewayErrorHandler.common_error_handler("create nat gateway")
@AWSRetry.jittered_backoff(catch_extra_error_codes=["InvalidElasticIpID.NotFound"])
def create_nat_gateway(
    client, **params: Dict[str, Union[str, bool, int, EC2TagSpecifications, List[str]]]
) -> Dict[str, Any]:
    return client.create_nat_gateway(**params)["NatGateway"]


# EC2 Elastic IP
class EC2ElasticIPErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidAddress.NotFound")


@EC2ElasticIPErrorHandler.list_error_handler("describe addresses", [])
@AWSRetry.jittered_backoff()
def describe_addresses(
    client, **params: Dict[str, Union[List[str], List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Union[str, List[Dict[str, str]]]]]:
    # The paginator does not exist for 'describe_addresses()'
    return client.describe_addresses(**params)["Addresses"]


@EC2ElasticIPErrorHandler.common_error_handler("release address")
@AWSRetry.jittered_backoff()
def release_address(client, allocation_id: str, network_border_group: Optional[str] = None) -> bool:
    params = {"AllocationId": allocation_id}
    if network_border_group:
        params["NetworkBorderGroup"] = network_border_group
    client.release_address(**params)
    return True


@EC2ElasticIPErrorHandler.common_error_handler("associate address")
@AWSRetry.jittered_backoff()
def associate_address(client, **params: Dict[str, Union[str, bool]]) -> Dict[str, str]:
    return client.associate_address(**params)


@EC2ElasticIPErrorHandler.common_error_handler("disassociate address")
@AWSRetry.jittered_backoff()
def disassociate_address(client, association_id: str) -> bool:
    client.disassociate_address(AssociationId=association_id)
    return True


@EC2ElasticIPErrorHandler.common_error_handler("allocate address")
@AWSRetry.jittered_backoff()
def allocate_address(client, **params: Dict[str, Union[str, EC2TagSpecifications]]) -> Dict[str, str]:
    return client.allocate_address(**params)


# EC2 VPC Endpoints
class EC2VpcEndpointsErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code(["InvalidVpcEndpoint.NotFound", "InvalidVpcEndpointId.NotFound"])


@EC2VpcEndpointsErrorHandler.list_error_handler("describe vpc endpoints", [])
@AWSRetry.jittered_backoff()
def describe_vpc_endpoints(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> Dict[str, Any]:
    paginator = client.get_paginator("describe_vpc_endpoints")
    return paginator.paginate(**params).build_full_result()["VpcEndpoints"]


@EC2VpcEndpointsErrorHandler.deletion_error_handler("delete vpc endpoints")
@AWSRetry.jittered_backoff()
def delete_vpc_endpoints(client, vpc_endpoint_ids: str) -> List[Dict[str, Union[str, Dict[str, str]]]]:
    result = client.delete_vpc_endpoints(VpcEndpointIds=vpc_endpoint_ids)
    return result.get("Unsuccessful", [])


@EC2ElasticIPErrorHandler.common_error_handler("create vpc endpoint")
@AWSRetry.jittered_backoff()
def create_vpc_endpoint(
    client,
    **params: Dict[
        str, Union[str, bool, List[str], Dict[str, Union[str, bool]], List[Dict[str, str]], EC2TagSpecifications]
    ],
) -> Dict[str, Any]:
    return client.create_vpc_endpoint(**params)["VpcEndpoint"]


# EC2 VPC Endpoint Services
class EC2VpcEndpointServiceErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidServiceName")


@EC2VpcEndpointServiceErrorHandler.list_error_handler("describe vpc endpoint services", default_value={})
@AWSRetry.jittered_backoff()
def describe_vpc_endpoint_services(
    client, filters: Optional[List[Dict[str, Any]]] = None, service_names: Optional[List[str]] = None
) -> Dict[str, Any]:
    """
    Wrap call to the AWS API describe_vpc_endpoint_services (used to describe available
    services to which you can create a VPC endpoint.)
        Parameters:
            client: The boto3 client.
            filters: Optional filters to pass to the API.
            service_names: the service names.
        Returns:
            results: A dictionnary with keys 'ServiceNames' and 'ServiceDetails'
    """
    paginator = client.get_paginator("describe_vpc_endpoint_services")
    params: dict[str, Any] = {}
    if filters:
        params["Filters"] = filters

    if service_names:
        params["ServiceNames"] = service_names

    results = paginator.paginate(**params).build_full_result()
    return results


# EC2 VPC DHCP Option
class EC2VpcDhcpOptionErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code(["InvalidDhcpOptionsID.NotFound", "InvalidDhcpOptionID.NotFound"])


@EC2VpcDhcpOptionErrorHandler.list_error_handler("describe dhcp options", [])
@AWSRetry.jittered_backoff()
def describe_dhcp_options(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_dhcp_options")
    return paginator.paginate(**params).build_full_result()["DhcpOptions"]


@EC2VpcDhcpOptionErrorHandler.deletion_error_handler("delete dhcp options")
@AWSRetry.jittered_backoff()
def delete_dhcp_options(client, dhcp_options_id: str) -> bool:
    client.delete_dhcp_options(DhcpOptionsId=dhcp_options_id)
    return True


@EC2VpcDhcpOptionErrorHandler.common_error_handler("associate dhcp options")
@AWSRetry.jittered_backoff()
def associate_dhcp_options(client, dhcp_options_id: str, vpc_id: str) -> bool:
    client.associate_dhcp_options(DhcpOptionsId=dhcp_options_id, VpcId=vpc_id)
    return True


@EC2VpcDhcpOptionErrorHandler.common_error_handler("create dhcp options")
@AWSRetry.jittered_backoff()
def create_dhcp_options(
    client, **params: Dict[str, Union[Dict[str, Union[str, List[str]]], EC2TagSpecifications]]
) -> Dict[str, Any]:
    return client.create_dhcp_options(**params)["DhcpOptions"]


# EC2 vpn Gateways
class EC2VpnGatewaysErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code(["InvalidVpnGatewayID.NotFound", "InvalidVpnGatewayState"])


@EC2VpnGatewaysErrorHandler.list_error_handler("describe vpn gateways", [])
@AWSRetry.jittered_backoff()
def describe_vpn_gateways(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    return client.describe_vpn_gateways(**params)["VpnGateways"]


@EC2VpnGatewaysErrorHandler.common_error_handler("create vpn gateway")
@AWSRetry.jittered_backoff(catch_extra_error_codes=["VpnGatewayLimitExceeded"])
def create_vpn_gateway(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> Dict[str, Any]:
    return client.create_vpn_gateway(**params)["VpnGateway"]


@EC2VpnGatewaysErrorHandler.deletion_error_handler("delete vpn gateway")
@AWSRetry.jittered_backoff()
def delete_vpn_gateway(client, vpn_gateway_id: str) -> bool:
    client.delete_vpn_gateway(VpnGatewayId=vpn_gateway_id)
    return True


@EC2VpnGatewaysErrorHandler.common_error_handler("attach vpn gateway")
@AWSRetry.jittered_backoff()
def attach_vpn_gateway(client, vpc_id: str, vpn_gateway_id: str) -> bool:
    client.attach_vpn_gateway(VpcId=vpc_id, VpnGatewayId=vpn_gateway_id)
    return True


@EC2VpnGatewaysErrorHandler.common_error_handler("detach vpn gateway")
@AWSRetry.jittered_backoff()
def detach_vpn_gateway(client, vpc_id: str, vpn_gateway_id: str) -> bool:
    client.detach_vpn_gateway(VpcId=vpc_id, VpnGatewayId=vpn_gateway_id)
    return True


# EC2 Volumes
class EC2VolumeErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidVolume.NotFound")


@EC2VolumeErrorHandler.list_error_handler("describe volumes", [])
@AWSRetry.jittered_backoff()
def describe_volumes(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_volumes")
    return paginator.paginate(**params).build_full_result()["Volumes"]


@EC2VolumeErrorHandler.deletion_error_handler("delete volume")
@AWSRetry.jittered_backoff()
def delete_volume(client, volume_id: str) -> bool:
    client.delete_volume(VolumeId=volume_id)
    return True


@EC2VolumeErrorHandler.common_error_handler("modify volume")
@AWSRetry.jittered_backoff()
def modify_volume(client, **params: Dict[str, Union[str, bool, int]]) -> Dict[str, Any]:
    return client.modify_volume(**params)["VolumeModification"]


@EC2VolumeErrorHandler.common_error_handler("modify volume")
@AWSRetry.jittered_backoff()
def create_volume(client, **params: Dict[str, Union[str, bool, int, EC2TagSpecifications]]) -> Dict[str, Any]:
    return client.create_volume(**params)


@EC2VolumeErrorHandler.common_error_handler("attach volume")
@AWSRetry.jittered_backoff()
def attach_volume(client, device: str, instance_id: str, volume_id: str) -> Dict[str, Any]:
    return client.attach_volume(Device=device, InstanceId=instance_id, VolumeId=volume_id)


@EC2VolumeErrorHandler.common_error_handler("attach volume")
@AWSRetry.jittered_backoff()
def detach_volume(client, volume_id: str, **params: Dict[str, Union[str, bool]]) -> Dict[str, Any]:
    return client.detach_volume(VolumeId=volume_id, **params)


# EC2 Instance
EC2_INSTANCE_CATCH_EXTRA_CODES = [
    "IncorrectState",
    "InsufficientInstanceCapacity",
    "InvalidInstanceID.NotFound",
]


class EC2InstanceErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidInstanceID.NotFound")


@EC2InstanceErrorHandler.list_error_handler("describe instances", [])
@AWSRetry.jittered_backoff()
def describe_instances(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_instances")
    return paginator.paginate(**params).build_full_result()["Reservations"]


@EC2InstanceErrorHandler.common_error_handler("modify instance attribute")
@AWSRetry.jittered_backoff(catch_extra_error_codes=EC2_INSTANCE_CATCH_EXTRA_CODES)
def modify_instance_attribute(
    client,
    instance_id: str,
    **params: Dict[
        str,
        Union[
            str,
            List[str],
            Dict[str, str],
            Dict[str, bool],
            Dict[str, bytes],
            Dict[str, Union[str, Dict[str, Union[str, bool]]]],
        ],
    ],
) -> bool:
    client.modify_instance_attribute(InstanceId=instance_id, **params)
    return True


@EC2InstanceErrorHandler.list_error_handler("terminate instances", [])
@AWSRetry.jittered_backoff()
def terminate_instances(client, instance_ids: List[str]) -> List[Dict[str, Any]]:
    return client.terminate_instances(InstanceIds=instance_ids)["TerminatingInstances"]


@EC2InstanceErrorHandler.list_error_handler("stop instances", [])
@AWSRetry.jittered_backoff()
def stop_instances(
    client, instance_ids: List[str], **params: Dict[str, Union[bool, List[str]]]
) -> List[Dict[str, Any]]:
    return client.stop_instances(InstanceIds=instance_ids, **params)["StoppingInstances"]


@EC2InstanceErrorHandler.list_error_handler("start instances", [])
@AWSRetry.jittered_backoff()
def start_instances(
    client, instance_ids: List[str], **params: Dict[str, Union[str, List[str]]]
) -> List[Dict[str, Any]]:
    return client.start_instances(InstanceIds=instance_ids, **params)["StartingInstances"]


@EC2InstanceErrorHandler.common_error_handler("run instances")
@AWSRetry.jittered_backoff(catch_extra_error_codes=EC2_INSTANCE_CATCH_EXTRA_CODES)
def run_instances(client, **params: Dict[str, Any]) -> Dict[str, Any]:
    return client.run_instances(**params)


@EC2InstanceErrorHandler.common_error_handler("describe instance attribute")
@AWSRetry.jittered_backoff(catch_extra_error_codes=EC2_INSTANCE_CATCH_EXTRA_CODES)
def describe_instance_attribute(client, instance_id: str, attribute: str) -> Dict[str, Any]:
    # The paginator does not exist for describe_instance_attribute()
    return client.describe_instance_attribute(InstanceId=instance_id, Attribute=attribute)


@EC2InstanceErrorHandler.common_error_handler("describe instance status")
@AWSRetry.jittered_backoff(catch_extra_error_codes=EC2_INSTANCE_CATCH_EXTRA_CODES)
def describe_instance_status(
    client, **params: Dict[str, Union[List[str], bool, int, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_instance_status")
    return paginator.paginate(**params).build_full_result()["InstanceStatuses"]


@EC2InstanceErrorHandler.common_error_handler("modify instance metadata options")
@AWSRetry.jittered_backoff(catch_extra_error_codes=EC2_INSTANCE_CATCH_EXTRA_CODES)
def modify_instance_metadata_options(
    client, instance_id: str, **params: Dict[str, Union[str, int]]
) -> Dict[str, Union[int, str]]:
    return client.modify_instance_metadata_options(InstanceId=instance_id, **params)["InstanceMetadataOptions"]


@EC2InstanceErrorHandler.common_error_handler("describe iam instance profile associations")
@AWSRetry.jittered_backoff(catch_extra_error_codes=EC2_INSTANCE_CATCH_EXTRA_CODES)
def describe_iam_instance_profile_associations(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> Dict[str, Any]:
    paginator = client.get_paginator("describe_iam_instance_profile_associations")
    return paginator.paginate(**params).build_full_result()["IamInstanceProfileAssociations"]


@EC2InstanceErrorHandler.common_error_handler("replace iam instance profile association")
@AWSRetry.jittered_backoff(catch_extra_error_codes=EC2_INSTANCE_CATCH_EXTRA_CODES)
def replace_iam_instance_profile_association(
    client, iam_instance_profile: Dict[str, str], association_id: str
) -> Dict[str, Union[int, str]]:
    return client.replace_iam_instance_profile_association(
        IamInstanceProfile=iam_instance_profile, AssociationId=association_id
    )["IamInstanceProfileAssociation"]


@EC2InstanceErrorHandler.common_error_handler("associate iam instance profile")
@AWSRetry.jittered_backoff(catch_extra_error_codes=EC2_INSTANCE_CATCH_EXTRA_CODES)
def associate_iam_instance_profile(client, iam_instance_profile: Dict[str, str], instance_id: str) -> Dict[str, Any]:
    return client.associate_iam_instance_profile(IamInstanceProfile=iam_instance_profile, InstanceId=instance_id)[
        "IamInstanceProfileAssociation"
    ]


# EC2 Instance Types
class EC2InstanceTypeErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    # @classmethod
    # def _is_missing(cls):
    #     return is_boto3_error_code("InvalidInstanceType")


@EC2InstanceTypeErrorHandler.list_error_handler("describe instance types", [])
@AWSRetry.jittered_backoff()
def describe_instance_types(
    client, **params: Dict[str, Union[List[str], Dict[str, Union[str, List[str]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_instance_types")
    return paginator.paginate(**params).build_full_result()["InstanceTypes"]


# EC2 Key
class EC2KeyErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidKeyPair.NotFound")


@EC2KeyErrorHandler.list_error_handler("describe key pairs", [])
@AWSRetry.jittered_backoff()
def describe_key_pairs(
    client, **params: Dict[str, Union[List[str], bool, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    # The paginator does not exist for `describe_key_pairs()`
    return client.describe_key_pairs(**params)["KeyPairs"]


@EC2KeyErrorHandler.common_error_handler("import key pair")
@AWSRetry.jittered_backoff()
def import_key_pair(
    client, **params: Dict[str, Union[str, bytes, EC2TagSpecifications]]
) -> Dict[str, Union[str, List[Dict[str, str]]]]:
    return client.import_key_pair(**params)


@EC2KeyErrorHandler.common_error_handler("create key pair")
@AWSRetry.jittered_backoff()
def create_key_pair(
    client, **params: Dict[str, Union[str, EC2TagSpecifications]]
) -> Dict[str, Union[str, List[Dict[str, str]]]]:
    return client.create_key_pair(**params)


@EC2KeyErrorHandler.deletion_error_handler("delete key pair")
@AWSRetry.jittered_backoff()
def delete_key_pair(client, key_name: Optional[str] = None, key_id: Optional[str] = None) -> bool:
    params = {}
    if key_name:
        params["KeyName"] = key_name
    if key_id:
        params["KeyPairId"] = key_id
    client.delete_key_pair(**params)
    return True


# EC2 Image
class EC2ImageErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidAMIID.Unavailable")


@EC2ImageErrorHandler.list_error_handler("describe images", [])
@AWSRetry.jittered_backoff()
def describe_images(
    client, **params: Dict[str, Union[List[str], bool, int, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    # 'DescribeImages' can be paginated depending on the boto3 version
    if client.can_paginate("describe_images"):
        paginator = client.get_paginator("describe_images")
        return paginator.paginate(**params).build_full_result()["Images"]
    else:
        return client.describe_images(**params)["Images"]


@EC2ImageErrorHandler.list_error_handler("describe image attribute", {})
@AWSRetry.jittered_backoff()
def describe_image_attribute(client, image_id: str, attribute: str) -> Optional[Dict[str, Any]]:
    # The paginator does not exist for `describe_image_attribute()`
    return client.describe_image_attribute(Attribute=attribute, ImageId=image_id)


@EC2ImageErrorHandler.deletion_error_handler("deregister image")
@AWSRetry.jittered_backoff()
def deregister_image(client, image_id: str) -> bool:
    client.deregister_image(ImageId=image_id)
    return True


@EC2ImageErrorHandler.common_error_handler("modify image attribute")
@AWSRetry.jittered_backoff()
def modify_image_attribute(client, image_id: str, **params: Dict[str, Any]) -> bool:
    client.modify_image_attribute(ImageId=image_id, **params)
    return True


@EC2ImageErrorHandler.common_error_handler("create image")
@AWSRetry.jittered_backoff()
def create_image(client, **params: Dict[str, Any]) -> Dict[str, str]:
    return client.create_image(**params)


@EC2ImageErrorHandler.common_error_handler("register image")
@AWSRetry.jittered_backoff()
def register_image(client, **params: Dict[str, Any]) -> Dict[str, str]:
    return client.register_image(**params)


# EC2 Snapshot
class EC2SnapshotErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidSnapshot.NotFound")


@EC2SnapshotErrorHandler.deletion_error_handler("delete snapshot")
@AWSRetry.jittered_backoff()
def delete_snapshot(client, snapshot_id: str) -> bool:
    client.delete_snapshot(SnapshotId=snapshot_id)
    return True


@EC2SnapshotErrorHandler.list_error_handler("describe snapshots", [])
@AWSRetry.jittered_backoff()
def describe_snapshots(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> Dict[str, Any]:
    # We do not use paginator here because the `ec2_snapshot_info` module excepts the NextToken to be returned
    return client.describe_snapshots(**params)


@EC2SnapshotErrorHandler.common_error_handler("describe snapshot attribute")
@AWSRetry.jittered_backoff()
def describe_snapshot_attribute(
    client, snapshot_id: str, attribute: str
) -> Dict[str, Union[str, List[Dict[str, str]]]]:
    return client.describe_snapshot_attribute(Attribute=attribute, SnapshotId=snapshot_id)


@EC2SnapshotErrorHandler.common_error_handler("reset snapshot attribute")
@AWSRetry.jittered_backoff()
def reset_snapshot_attribute(client, snapshot_id: str, attribute: str) -> bool:
    client.reset_snapshot_attribute(Attribute=attribute, SnapshotId=snapshot_id)
    return True


@EC2SnapshotErrorHandler.common_error_handler("modify snapshot attribute")
@AWSRetry.jittered_backoff()
def modify_snapshot_attribute(client, snapshot_id: str, **params: Dict[str, Any]) -> bool:
    client.modify_snapshot_attribute(SnapshotId=snapshot_id, **params)
    return True


@EC2SnapshotErrorHandler.common_error_handler("create snapshot")
# Handle SnapshotCreationPerVolumeRateExceeded separately because we need a much
# longer delay than normal
@AWSRetry.jittered_backoff(catch_extra_error_codes=["SnapshotCreationPerVolumeRateExceeded"], delay=15)
def create_snapshot(client, volume_id: str, **params: Dict[str, Any]) -> Dict[str, Any]:
    return client.create_snapshot(VolumeId=volume_id, **params)


# EC2 ENI
class EC2NetworkInterfacesErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidNetworkInterfaceID.NotFound")


@EC2NetworkInterfacesErrorHandler.list_error_handler("describe network interfaces", [])
@AWSRetry.jittered_backoff()
def describe_network_interfaces(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_network_interfaces")
    return paginator.paginate(**params).build_full_result()["NetworkInterfaces"]


@EC2NetworkInterfacesErrorHandler.deletion_error_handler("delete network interface")
@AWSRetry.jittered_backoff()
def delete_network_interface(client, network_interface_id: str) -> bool:
    client.delete_network_interface(NetworkInterfaceId=network_interface_id)
    return True


@EC2NetworkInterfacesErrorHandler.common_error_handler("create network interface")
@AWSRetry.jittered_backoff()
def create_network_interface(client, **params: Dict[str, Any]) -> Dict[str, str]:
    return client.create_network_interface(**params)["NetworkInterface"]


@EC2NetworkInterfacesErrorHandler.common_error_handler("attach network interface")
@AWSRetry.jittered_backoff()
def attach_network_interface(
    client, **params: Dict[str, Union[str, int, Dict[str, Union[bool, Dict[str, bool]]]]]
) -> Dict[str, Union[str, int]]:
    return client.attach_network_interface(**params)


@EC2NetworkInterfacesErrorHandler.common_error_handler("detach network interface")
@AWSRetry.jittered_backoff()
def detach_network_interface(client, attachment_id: str, force: Optional[bool] = None) -> bool:
    params = {"AttachmentId": attachment_id}
    if force is not None:
        params["Force"] = force
    client.detach_network_interface(**params)
    return True


@EC2NetworkInterfacesErrorHandler.common_error_handler("assign private ip addresses")
@AWSRetry.jittered_backoff()
def assign_private_ip_addresses(
    client, **params: Dict[str, Union[bool, int, str, List[str]]]
) -> Dict[str, Union[str, List[Dict[str, str]]]]:
    return client.assign_private_ip_addresses(**params)


@EC2NetworkInterfacesErrorHandler.common_error_handler("unassign private ip addresses")
@AWSRetry.jittered_backoff()
def unassign_private_ip_addresses(client, **params: Dict[str, Union[str, List[str]]]) -> bool:
    client.unassign_private_ip_addresses(**params)
    return True


@EC2NetworkInterfacesErrorHandler.common_error_handler("modify network interface attribute")
@AWSRetry.jittered_backoff()
def modify_network_interface_attribute(client, **params: Dict[str, Any]) -> bool:
    client.modify_network_interface_attribute(**params)
    return True


# EC2 Import Image
class EC2ImportImageErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidConversionTaskId")


@EC2ImportImageErrorHandler.list_error_handler("describe import image tasks", [])
@AWSRetry.jittered_backoff()
def describe_import_image_tasks(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_import_image_tasks")
    return paginator.paginate(**params).build_full_result()["ImportImageTasks"]


def describe_import_image_tasks_as_snake_dict(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    result = []
    for import_image_info in describe_import_image_tasks(client, **params):
        image = copy.deepcopy(import_image_info)
        image["Tags"] = boto3_tag_list_to_ansible_dict(image["Tags"])
        result.append(camel_dict_to_snake_dict(image, ignore_list=["Tags"]))
    return result


@EC2ImportImageErrorHandler.deletion_error_handler("cancel import task")
@AWSRetry.jittered_backoff()
def cancel_import_task(client, import_task_id: str, cancel_reason: Optional[str] = None) -> bool:
    params = {"ImportTaskId": import_task_id}
    if cancel_reason:
        params["CancelReason"] = cancel_reason
    client.cancel_import_task(**params)
    return True


@EC2ImportImageErrorHandler.common_error_handler("import image")
@AWSRetry.jittered_backoff()
def import_image(client, **params: Dict[str, Any]) -> Dict[str, Any]:
    return client.import_image(**params)


# EC2 Spot instance
class EC2SpotInstanceRequestErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidSpotInstanceRequestID.NotFound")


@EC2SpotInstanceRequestErrorHandler.list_error_handler("describe spot instance requests", [])
@AWSRetry.jittered_backoff()
def describe_spot_instance_requests(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_spot_instance_requests")
    return paginator.paginate(**params).build_full_result()["SpotInstanceRequests"]


@EC2SpotInstanceRequestErrorHandler.common_error_handler("cancel spot instance requests")
@AWSRetry.jittered_backoff()
def cancel_spot_instance_requests(client, spot_instance_request_ids: List[str]) -> List[Dict[str, str]]:
    return client.cancel_spot_instance_requests(SpotInstanceRequestIds=spot_instance_request_ids)[
        "CancelledSpotInstanceRequests"
    ]


@EC2SpotInstanceRequestErrorHandler.common_error_handler("request spot instances")
@AWSRetry.jittered_backoff()
def request_spot_instances(client, **params: Dict[str, Any]) -> List[Dict[str, Any]]:
    return client.request_spot_instances(**params)["SpotInstanceRequests"]


# EC2 Security Groups
class EC2SecurityGroupsErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidGroup.NotFound")


@EC2SecurityGroupsErrorHandler.list_error_handler("describe security groups", [])
@AWSRetry.jittered_backoff()
def describe_security_groups(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_security_groups")
    return paginator.paginate(**params).build_full_result()["SecurityGroups"]


@EC2SecurityGroupsErrorHandler.deletion_error_handler("delete security group")
@AWSRetry.jittered_backoff()
def delete_security_group(client, group_id: Optional[str], group_name: Optional[str] = None) -> bool:
    params = {}
    if group_id:
        params["GroupId"] = group_id
    if group_name:
        params["GroupName"] = group_name
    client.delete_security_group(**params)
    return True


@EC2SecurityGroupsErrorHandler.common_error_handler("create security group")
@AWSRetry.jittered_backoff()
def create_security_group(client, **params: Dict[str, Union[str, EC2TagSpecifications]]) -> Dict[str, Any]:
    return client.create_security_group(**params)


@EC2SecurityGroupsErrorHandler.common_error_handler("update security group rule descriptions egress")
@AWSRetry.jittered_backoff()
def update_security_group_rule_descriptions_egress(client, **params: Dict[str, Any]) -> bool:
    return client.update_security_group_rule_descriptions_egress(**params)["Return"]


@EC2SecurityGroupsErrorHandler.common_error_handler("update security group rule descriptions ingress")
@AWSRetry.jittered_backoff()
def update_security_group_rule_descriptions_ingress(client, **params: Dict[str, Any]) -> bool:
    return client.update_security_group_rule_descriptions_ingress(**params)["Return"]


@EC2SecurityGroupsErrorHandler.common_error_handler("revoke security group ingress")
@AWSRetry.jittered_backoff()
def revoke_security_group_ingress(client, **params: Dict[str, Any]) -> bool:
    return client.revoke_security_group_ingress(**params)["Return"]


@EC2SecurityGroupsErrorHandler.common_error_handler("revoke security group egress")
@AWSRetry.jittered_backoff()
def revoke_security_group_egress(client, **params: Dict[str, Any]) -> bool:
    return client.revoke_security_group_egress(**params)["Return"]


@EC2SecurityGroupsErrorHandler.common_error_handler("authorize security group ingress")
@AWSRetry.jittered_backoff()
def authorize_security_group_ingress(client, **params: Dict[str, Any]) -> bool:
    return client.authorize_security_group_ingress(**params)["Return"]


@EC2SecurityGroupsErrorHandler.common_error_handler("authorize security group egress")
@AWSRetry.jittered_backoff()
def authorize_security_group_egress(client, **params: Dict[str, Any]) -> bool:
    return client.authorize_security_group_egress(**params)["Return"]


# EC2 Egress only internet Gateway
class EC2EgressOnlyInternetGatewayErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidGatewayID.NotFound")


@EC2EgressOnlyInternetGatewayErrorHandler.list_error_handler("describe egress only internet gateways", [])
@AWSRetry.jittered_backoff()
def describe_egress_only_internet_gateways(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_egress_only_internet_gateways")
    return paginator.paginate(**params).build_full_result()["EgressOnlyInternetGateways"]


@EC2EgressOnlyInternetGatewayErrorHandler.deletion_error_handler("delete egress only internet gateway")
@AWSRetry.jittered_backoff()
def delete_egress_only_internet_gateway(client, egress_only_internet_gateway_id: str) -> bool:
    return client.delete_egress_only_internet_gateway(EgressOnlyInternetGatewayId=egress_only_internet_gateway_id)[
        "ReturnCode"
    ]


@EC2EgressOnlyInternetGatewayErrorHandler.common_error_handler("create egress only internet gateway")
@AWSRetry.jittered_backoff()
def create_egress_only_internet_gateway(
    client, vpc_id: str, tags: Optional[EC2TagSpecifications] = None
) -> Dict[str, Any]:
    params = {"VpcId": vpc_id}
    if tags:
        params["TagSpecifications"] = boto3_tag_specifications(tags, types="egress-only-internet-gateway")
    return client.create_egress_only_internet_gateway(**params)


# EC2 Network ACL
class EC2NetworkAclErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidNetworkAclID.NotFound")


@EC2NetworkAclErrorHandler.list_error_handler("describe network acls", [])
@AWSRetry.jittered_backoff()
def describe_network_acls(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_network_acls")
    return paginator.paginate(**params).build_full_result()["NetworkAcls"]


@EC2NetworkAclErrorHandler.common_error_handler("create network acl")
@AWSRetry.jittered_backoff()
def create_network_acl(client, vpc_id: str, tags: Optional[EC2TagSpecifications] = None) -> Dict[str, Any]:
    params = {"VpcId": vpc_id}
    if tags:
        params["TagSpecifications"] = boto3_tag_specifications(tags, types="network-acl")
    return client.create_network_acl(**params)["NetworkAcl"]


@EC2NetworkAclErrorHandler.common_error_handler("create network acl entry")
@AWSRetry.jittered_backoff()
def create_network_acl_entry(
    client,
    network_acl_id: str,
    protocol: str,
    egress: bool,
    rule_action: str,
    rule_number: int,
    **params: Dict[str, Any],
) -> bool:
    args = {
        "Egress": egress,
        "NetworkAclId": network_acl_id,
        "Protocol": protocol,
        "RuleAction": rule_action,
        "RuleNumber": rule_number,
    }
    if params:
        args.update(params)
    client.create_network_acl_entry(**args)
    return True


@EC2NetworkAclErrorHandler.deletion_error_handler("delete network acl")
@AWSRetry.jittered_backoff()
def delete_network_acl(client, network_acl_id: str) -> bool:
    client.delete_network_acl(NetworkAclId=network_acl_id)
    return True


@EC2NetworkAclErrorHandler.deletion_error_handler("delete network acl entry")
@AWSRetry.jittered_backoff()
def delete_network_acl_entry(client, network_acl_id: str, rule_number: int, egress: bool) -> bool:
    client.delete_network_acl_entry(NetworkAclId=network_acl_id, Egress=egress, RuleNumber=rule_number)
    return True


@EC2NetworkAclErrorHandler.common_error_handler("replace network acl entry")
@AWSRetry.jittered_backoff()
def replace_network_acl_entry(
    client,
    network_acl_id: str,
    protocol: str,
    egress: bool,
    rule_action: str,
    rule_number: int,
    **params: Dict[str, Any],
) -> Dict[str, Any]:
    args = {
        "Egress": egress,
        "NetworkAclId": network_acl_id,
        "Protocol": protocol,
        "RuleAction": rule_action,
        "RuleNumber": rule_number,
    }
    if params:
        args.update(params)
    return client.replace_network_acl_entry(**args)


@EC2NetworkAclErrorHandler.common_error_handler("replace network acl association")
@AWSRetry.jittered_backoff()
def replace_network_acl_association(client, network_acl_id: str, association_id: str) -> str:
    return client.replace_network_acl_association(NetworkAclId=network_acl_id, AssociationId=association_id)[
        "NewAssociationId"
    ]


# EC2 Placement Group
class EC2PlacementGroupErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidPlacementGroup.Unknown")


@EC2PlacementGroupErrorHandler.list_error_handler("describe placement group", [])
@AWSRetry.jittered_backoff()
def describe_ec2_placement_groups(
    client, **params: Dict[str, Union[List[str], int, List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    return client.describe_placement_groups(**params)["PlacementGroups"]


@EC2PlacementGroupErrorHandler.deletion_error_handler("delete placement group")
@AWSRetry.jittered_backoff()
def delete_ec2_placement_group(client, group_name: str) -> bool:
    client.delete_placement_group(GroupName=group_name)
    return True


@EC2PlacementGroupErrorHandler.common_error_handler("create placement group")
@AWSRetry.jittered_backoff()
def create_ec2_placement_group(client, **params: Dict[str, Union[str, EC2TagSpecifications]]) -> Dict[str, Any]:
    return client.create_placement_group(**params)["PlacementGroup"]


# EC2 Launch template
class EC2LaunchTemplateErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code(["InvalidLaunchTemplateName.NotFoundException", "InvalidLaunchTemplateId.NotFound"])


@EC2LaunchTemplateErrorHandler.list_error_handler("describe launch templates", [])
@AWSRetry.jittered_backoff()
def describe_launch_templates(
    client,
    launch_template_ids: Optional[List[str]] = None,
    launch_template_names: Optional[List[str]] = None,
    filters: Optional[List[Dict[str, List[str]]]] = None,
) -> List[Dict[str, Any]]:
    params = {}
    if launch_template_ids:
        params["LaunchTemplateIds"] = launch_template_ids
    if launch_template_names:
        params["LaunchTemplateNames"] = launch_template_names
    if filters:
        params["Filters"] = filters
    paginator = client.get_paginator("describe_launch_templates")
    return paginator.paginate(**params).build_full_result()["LaunchTemplates"]


@EC2LaunchTemplateErrorHandler.common_error_handler("describe launch template versions")
@AWSRetry.jittered_backoff()
def describe_launch_template_versions(client, **params: Dict[str, Any]) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_launch_template_versions")
    return paginator.paginate(**params).build_full_result()["LaunchTemplateVersions"]


@EC2LaunchTemplateErrorHandler.common_error_handler("delete launch template versions")
@AWSRetry.jittered_backoff()
def delete_launch_template_versions(
    client, versions: List[str], launch_template_id: Optional[str] = None, launch_template_name: Optional[str] = None
) -> Dict[str, Any]:
    params = {}
    if launch_template_id:
        params["LaunchTemplateId"] = launch_template_id
    if launch_template_name:
        params["LaunchTemplateName"] = launch_template_name
    response = {
        "UnsuccessfullyDeletedLaunchTemplateVersions": [],
        "SuccessfullyDeletedLaunchTemplateVersions": [],
    }
    # Using this API, You can specify up to 200 launch template version numbers.
    for i in range(0, len(versions), 200):
        result = client.delete_launch_template_versions(Versions=list(versions[i : i + 200]), **params)
        for x in ("SuccessfullyDeletedLaunchTemplateVersions", "UnsuccessfullyDeletedLaunchTemplateVersions"):
            response[x] += result.get(x, [])
    return response


@EC2LaunchTemplateErrorHandler.common_error_handler("delete launch template")
@AWSRetry.jittered_backoff()
def delete_launch_template(
    client, launch_template_id: Optional[str] = None, launch_template_name: Optional[str] = None
) -> Dict[str, Any]:
    params = {}
    if launch_template_id:
        params["LaunchTemplateId"] = launch_template_id
    if launch_template_name:
        params["LaunchTemplateName"] = launch_template_name
    return client.delete_launch_template(**params)["LaunchTemplate"]


@EC2LaunchTemplateErrorHandler.common_error_handler("create launch template")
@AWSRetry.jittered_backoff()
def create_launch_template(
    client,
    launch_template_name: str,
    launch_template_data: Dict[str, Any],
    tags: Optional[EC2TagSpecifications] = None,
    **kwargs: Dict[str, Any],
) -> Dict[str, Any]:
    params = {"LaunchTemplateName": launch_template_name, "LaunchTemplateData": launch_template_data}
    if tags:
        params["TagSpecifications"] = boto3_tag_specifications(tags, types="launch-template")
    params.update(kwargs)
    return client.create_launch_template(**params)["LaunchTemplate"]


@EC2LaunchTemplateErrorHandler.common_error_handler("create launch template version")
@AWSRetry.jittered_backoff()
def create_launch_template_version(
    client, launch_template_data: Dict[str, Any], **params: Dict[str, Any]
) -> Dict[str, Any]:
    return client.create_launch_template_version(LaunchTemplateData=launch_template_data, **params)[
        "LaunchTemplateVersion"
    ]


@EC2LaunchTemplateErrorHandler.common_error_handler("modify launch template")
@AWSRetry.jittered_backoff()
def modify_launch_template(client, **params: Dict[str, Any]) -> Dict[str, Any]:
    return client.modify_launch_template(**params)["LaunchTemplate"]


def get_ec2_security_group_ids_from_names(sec_group_list, ec2_connection, vpc_id=None, boto3=None):
    """Return list of security group IDs from security group names. Note that security group names are not unique
    across VPCs.  If a name exists across multiple VPCs and no VPC ID is supplied, all matching IDs will be returned. This
    will probably lead to a boto exception if you attempt to assign both IDs to a resource so ensure you wrap the call in
    a try block
    """

    def get_sg_name(sg):
        return str(sg["GroupName"])

    def get_sg_id(sg):
        return str(sg["GroupId"])

    sec_group_id_list = []

    if isinstance(sec_group_list, str):
        sec_group_list = [sec_group_list]

    # Get all security groups
    if vpc_id:
        filters = [
            {
                "Name": "vpc-id",
                "Values": [
                    vpc_id,
                ],
            }
        ]
        all_sec_groups = ec2_connection.describe_security_groups(Filters=filters)["SecurityGroups"]
    else:
        all_sec_groups = ec2_connection.describe_security_groups()["SecurityGroups"]

    unmatched = set(sec_group_list).difference(str(get_sg_name(all_sg)) for all_sg in all_sec_groups)
    sec_group_name_list = list(set(sec_group_list) - set(unmatched))

    if len(unmatched) > 0:
        # If we have unmatched names that look like an ID, assume they are
        sec_group_id_list[:] = [sg for sg in unmatched if re.match("sg-[a-fA-F0-9]+$", sg)]
        still_unmatched = [sg for sg in unmatched if not re.match("sg-[a-fA-F0-9]+$", sg)]
        if len(still_unmatched) > 0:
            raise ValueError(f"The following group names are not valid: {', '.join(still_unmatched)}")

    sec_group_id_list += [get_sg_id(all_sg) for all_sg in all_sec_groups if get_sg_name(all_sg) in sec_group_name_list]

    return sec_group_id_list


# EC2 Transit Gateway VPC Attachment Error handler
class EC2TransitGatewayVPCAttachmentErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidGatewayID.NotFound")


@EC2TransitGatewayVPCAttachmentErrorHandler.common_error_handler("describe transit gateway vpc attachments")
@AWSRetry.jittered_backoff()
def describe_transit_gateway_vpc_attachments(
    client, **params: Dict[str, Union[List[str], bool, List[Dict[str, Union[str, List[str]]]]]]
) -> List:
    paginator = client.get_paginator("describe_transit_gateway_vpc_attachments")
    return paginator.paginate(**params).build_full_result()["TransitGatewayVpcAttachments"]


@EC2TransitGatewayVPCAttachmentErrorHandler.common_error_handler("create transit gateway vpc attachment")
@AWSRetry.jittered_backoff()
def create_transit_gateway_vpc_attachment(
    client, **params: Dict[str, Union[List[str], bool, List[Dict[str, Union[str, List[str]]]]]]
) -> Dict[str, Any]:
    return client.create_transit_gateway_vpc_attachment(**params)["TransitGatewayVpcAttachment"]


@EC2TransitGatewayVPCAttachmentErrorHandler.common_error_handler("modify transit gateway vpc attachment")
@AWSRetry.jittered_backoff()
def modify_transit_gateway_vpc_attachment(
    client, **params: Dict[str, Union[List[str], bool, List[Dict[str, Union[str, List[str]]]]]]
) -> Dict[str, Any]:
    return client.modify_transit_gateway_vpc_attachment(**params)["TransitGatewayVpcAttachment"]


@EC2TransitGatewayVPCAttachmentErrorHandler.deletion_error_handler("delete transit gateway vpc attachment")
@AWSRetry.jittered_backoff()
def delete_transit_gateway_vpc_attachment(client, transit_gateway_attachment_id: str) -> bool:
    client.delete_transit_gateway_vpc_attachment(TransitGatewayAttachmentId=transit_gateway_attachment_id)[
        "TransitGatewayVpcAttachment"
    ]
    return True


def add_ec2_tags(client, module, resource_id, tags_to_set, retry_codes=None):
    """
    Sets Tags on an EC2 resource.

    :param client: an EC2 boto3 client
    :param module: an AnsibleAWSModule object
    :param resource_id: the identifier for the resource
    :param tags_to_set: A dictionary of key/value pairs to set
    :param retry_codes: additional boto3 error codes to trigger retries
    """

    if not tags_to_set:
        return False
    if module.check_mode:
        return True

    if not retry_codes:
        retry_codes = []

    try:
        tags_to_add = ansible_dict_to_boto3_tag_list(tags_to_set)
        AWSRetry.jittered_backoff(retries=10, catch_extra_error_codes=retry_codes)(client.create_tags)(
            Resources=[resource_id], Tags=tags_to_add
        )
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(e, msg=f"Unable to add tags {tags_to_set} to {resource_id}")
    return True


def remove_ec2_tags(client, module, resource_id, tags_to_unset, retry_codes=None):
    """
    Removes Tags from an EC2 resource.

    :param client: an EC2 boto3 client
    :param module: an AnsibleAWSModule object
    :param resource_id: the identifier for the resource
    :param tags_to_unset: a list of tag keys to removes
    :param retry_codes: additional boto3 error codes to trigger retries
    """

    if not tags_to_unset:
        return False
    if module.check_mode:
        return True

    if not retry_codes:
        retry_codes = []

    tags_to_remove = [dict(Key=tagkey) for tagkey in tags_to_unset]

    try:
        AWSRetry.jittered_backoff(retries=10, catch_extra_error_codes=retry_codes)(client.delete_tags)(
            Resources=[resource_id], Tags=tags_to_remove
        )
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(e, msg=f"Unable to delete tags {tags_to_unset} from {resource_id}")
    return True


def describe_ec2_tags(client, module, resource_id, resource_type=None, retry_codes=None):
    """
    Performs a paginated search of EC2 resource tags.

    :param client: an EC2 boto3 client
    :param module: an AnsibleAWSModule object
    :param resource_id: the identifier for the resource
    :param resource_type: the type of the resource
    :param retry_codes: additional boto3 error codes to trigger retries
    """
    filters = {"resource-id": resource_id}
    if resource_type:
        filters["resource-type"] = resource_type
    filters = ansible_dict_to_boto3_filter_list(filters)

    if not retry_codes:
        retry_codes = []

    try:
        retry_decorator = AWSRetry.jittered_backoff(retries=10, catch_extra_error_codes=retry_codes)
        results = paginated_query_with_retries(
            client, "describe_tags", retry_decorator=retry_decorator, Filters=filters
        )
        return boto3_tag_list_to_ansible_dict(results.get("Tags", None))
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(e, msg=f"Failed to describe tags for EC2 Resource: {resource_id}")


def ensure_ec2_tags(client, module, resource_id, resource_type=None, tags=None, purge_tags=True, retry_codes=None):
    """
    Updates the tags on an EC2 resource.

    To remove all tags the tags parameter must be explicitly set to an empty dictionary.

    :param client: an EC2 boto3 client
    :param module: an AnsibleAWSModule object
    :param resource_id: the identifier for the resource
    :param resource_type: the type of the resource
    :param tags: the Tags to apply to the resource
    :param purge_tags: whether tags missing from the tag list should be removed
    :param retry_codes: additional boto3 error codes to trigger retries
    :return: changed: returns True if the tags are changed
    """

    if tags is None:
        return False

    if not retry_codes:
        retry_codes = []

    changed = False
    current_tags = describe_ec2_tags(client, module, resource_id, resource_type, retry_codes)

    tags_to_set, tags_to_unset = compare_aws_tags(current_tags, tags, purge_tags)

    if purge_tags and not tags:
        tags_to_unset = current_tags

    changed |= remove_ec2_tags(client, module, resource_id, tags_to_unset, retry_codes)
    changed |= add_ec2_tags(client, module, resource_id, tags_to_set, retry_codes)

    return changed


def normalize_ec2_vpc_dhcp_config(option_config: List[Dict[str, Any]]) -> Dict[str, Any]:
    """
    The boto2 module returned a config dict, but boto3 returns a list of dicts
    Make the data we return look like the old way, so we don't break users.
    This is also much more user-friendly.
    boto3:
        'DhcpConfigurations': [
            {'Key': 'domain-name', 'Values': [{'Value': 'us-west-2.compute.internal'}]},
            {'Key': 'domain-name-servers', 'Values': [{'Value': 'AmazonProvidedDNS'}]},
            {'Key': 'netbios-name-servers', 'Values': [{'Value': '1.2.3.4'}, {'Value': '5.6.7.8'}]},
            {'Key': 'netbios-node-type', 'Values': [1]},
            {'Key': 'ntp-servers', 'Values': [{'Value': '1.2.3.4'}, {'Value': '5.6.7.8'}]}
        ],
    The module historically returned:
        "new_options": {
            "domain-name": "ec2.internal",
            "domain-name-servers": ["AmazonProvidedDNS"],
            "netbios-name-servers": ["10.0.0.1", "10.0.1.1"],
            "netbios-node-type": "1",
            "ntp-servers": ["10.0.0.2", "10.0.1.2"]
        },
    """
    config_data = {}

    if len(option_config) == 0:
        # If there is no provided config, return the empty dictionary
        return config_data

    for config_item in option_config:
        # Handle single value keys
        if config_item["Key"] == "netbios-node-type":
            if isinstance(config_item["Values"], int):
                config_data["netbios-node-type"] = str((config_item["Values"]))
            elif isinstance(config_item["Values"], list):
                config_data["netbios-node-type"] = str((config_item["Values"][0]["Value"]))
        # Handle actual lists of values
        for option in ["domain-name", "domain-name-servers", "ntp-servers", "netbios-name-servers"]:
            if config_item["Key"] == option:
                config_data[option] = [val["Value"] for val in config_item["Values"]]

    return config_data


def determine_iam_arn_from_name(iam_client, name_or_arn: str) -> str:
    if validate_aws_arn(name_or_arn, service="iam", resource_type="instance-profile"):
        return name_or_arn

    iam_instance_profiles = list_iam_instance_profiles(iam_client, name=name_or_arn)
    if not iam_instance_profiles:
        raise AnsibleEC2Error(message=f"Could not find IAM instance profile {name_or_arn}")
    return iam_instance_profiles[0]["Arn"]


# EC2 Transit Gateway
class EC2TransitGatewayErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidTransitGatewayID.NotFound")


@EC2TransitGatewayErrorHandler.list_error_handler("describe transit gateway", [])
@AWSRetry.jittered_backoff()
def describe_ec2_transit_gateways(
    client, **params: Dict[str, Union[List[str], List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_transit_gateways")
    return paginator.paginate(**params).build_full_result()["TransitGateways"]


@EC2TransitGatewayErrorHandler.common_error_handler("create transit gateway")
@AWSRetry.jittered_backoff()
def create_ec2_transit_gateway(
    client, **params: Dict[str, Union[List[str], List[Dict[str, Union[str, List[str]]]]]]
) -> Dict[str, Any]:
    return client.create_transit_gateway(**params)["TransitGateway"]


@EC2TransitGatewayErrorHandler.deletion_error_handler("delete transit gateway")
@AWSRetry.jittered_backoff()
def delete_ec2_transit_gateway(client, transit_gateway_id: str) -> bool:
    client.delete_transit_gateway(TransitGatewayId=transit_gateway_id)
    return True


# EC2 Dedicated host
class EC2DedicatedHost(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleEC2Error

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("InvalidHostID.NotFound")


@EC2DedicatedHost.list_error_handler("describe dedicated host", [])
@AWSRetry.jittered_backoff()
def describe_ec2_dedicated_hosts(
    client, **params: Dict[str, Union[List[str], List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_hosts")
    return paginator.paginate(**params).build_full_result()["Hosts"]


@EC2DedicatedHost.list_error_handler("describe mac dedicated host", [])
@AWSRetry.jittered_backoff()
def describe_ec2_mac_dedicated_hosts(
    client, **params: Dict[str, Union[List[str], List[Dict[str, Union[str, List[str]]]]]]
) -> List[Dict[str, Any]]:
    paginator = client.get_paginator("describe_mac_hosts")
    return paginator.paginate(**params).build_full_result()["MacHosts"]


@EC2DedicatedHost.deletion_error_handler("release dedicated host")
@AWSRetry.jittered_backoff()
def release_ec2_dedicated_host(client, host_id: Union[str, List[str]]) -> bool:
    host_ids = host_id
    if isinstance(host_id, str):
        host_ids = [host_id]
    client.release_hosts(HostIds=host_ids)
    return True


@EC2DedicatedHost.common_error_handler("allocate dedicated hosts")
@AWSRetry.jittered_backoff()
def allocate_ec2_dedicated_hosts(
    client, availability_zone: str, **params: Dict[str, Union[List[str], List[Dict[str, Union[str, List[str]]]]]]
) -> List[str]:
    return client.allocate_hosts(AvailabilityZone=availability_zone, **params)["HostIds"]


@EC2DedicatedHost.common_error_handler("modify dedicated hosts")
@AWSRetry.jittered_backoff()
def modify_ec2_dedicated_hosts(
    client,
    host_id: Union[List[str], str],
    **params: Dict[str, Union[List[str], List[Dict[str, Union[str, List[str]]]]]],
) -> Dict[str, Any]:
    host_ids = host_id
    if isinstance(host_id, str):
        host_ids = [host_id]
    return client.modify_hosts(HostIds=host_ids, **params)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2024, Andrew Klychkov (@Andersson007) <andrew.a.klychkov@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = r'''
---
module: clickhouse_info

short_description: Gather ClickHouse server information using the clickhouse-driver Client interface

description:
  - Gather ClickHouse server information using the
    L(clickhouse-driver,https://clickhouse-driver.readthedocs.io/en/latest) Client interface.
  - Does not change server state.

attributes:
  check_mode:
    description: Supports check_mode.
    support: full
  idempotent:
    description: Module only reads info. Always returns changed=false.
    support: full

version_added: '0.1.0'

author:
  - Andrew Klychkov (@Andersson007)
  - Aleksandr Vagachev (@aleksvagachev)

extends_documentation_fragment:
  - community.clickhouse.client_inst_opts

options:
  limit:
    description:
      - Limits a set of return values you want to get.
      - See the Return section for acceptable values.
      - If not specified, returns all supported return values.
    type: list
    elements: str
    version_added: '0.2.0'
'''

EXAMPLES = r'''
- name: Get server information
  register: result
  community.clickhouse.clickhouse_info:
    login_host: localhost
    login_user: alice
    login_db: foo
    login_password: my_password

- name: Print returned data
  ansible.builtin.debug:
    var: result

- name: Limit return values with users and roles
  register: result
  community.clickhouse.clickhouse_info:
    limit:
      - users
      - roles
'''

# When adding new ret values,
# please add it to the ret_val_func_mapping dictionary!
RETURN = r'''
driver:
  description: Python driver information.
  returned: success
  type: dict
  sample: { "version": "0.2.6" }
version:
  description: Clickhouse server version.
  returned: success
  type: dict
  sample: {"raw": "23.12.2.59", "year": 23, "feature": 12, "maintenance": 2, "build": 59, "type": null }
databases:
  description:
    - The content of the system.databases table with names as keys.
    - Be sure your I(login_user) has permissions.
  returned: success
  type: dict
  sample: { "system": "..." }
users:
  description:
    - The content of the system.users table with names as keys.
    - Be sure your I(login_user) has permissions.
  returned: success
  type: dict
  sample: { "default": "..." }
roles:
  description:
    - The content of the system.roles table with names as keys.
    - Be sure your I(login_user) has permissions.
  returned: success
  type: dict
  sample: { "accountant": "..." }
settings:
  description:
    - The content of the system.settings table with names as keys.
  returned: success
  type: dict
  sample: { "zstd_window_log_max": "..." }
clusters:
  description:
    - The content of the system.clusters table with names as keys.
  returned: success
  type: dict
  sample: { "test_cluster_two_shards": "..." }
tables:
  description:
    - The content of the system.tables table with the table names as keys.
  returned: success
  type: dict
  sample: { "system": { "settings": "..." } }
  version_added: '0.3.0'
merge_tree_settings:
  description:
    - The content of the system.merge_tree_settings table with names as keys.
  returned: success
  type: dict
  sample: { "merge_max_block_size": "..." }
  version_added: '0.3.0'
dictionaries:
  description:
    - The content of the system.dictionaries table with dictionary names as keys.
  returned: success
  type: dict
  sample: { "database": { "dictionary": "..." } }
  version_added: '0.3.0'
quotas:
  description:
    - The content of the system.quotas table with quota names as keys.
  returned: success
  type: dict
  sample: { "default": "..." }
  version_added: '0.4.0'
settings_profiles:
  description:
    - The content of the system.settings_profiles table with profile names as keys.
  returned: success
  type: dict
  sample: { "readonly": "..." }
  version_added: '0.4.0'
storage_policies:
  description:
    - The content of the system.storage_policies table with storage_policies names as keys.
  returned: success
  type: dict
  sample: { "storage_policies": "..." }
  version_added: '0.4.0'
grants:
  description:
    - The content of the system.grants table for users and roles as keys.
  returned: success
  type: dict
  sample: { "roles": {"..."}, "users": {"..."} }
  version_added: '0.7.0'
settings_profile_elements:
  description:
    - The content of the system.settings_profile_elements table for users, roles, profiles as keys.
  returned: success
  type: dict
  sample: { "roles": {"..."}, "users": {"..."}, "profiles": {"..."} }
  version_added: '0.7.0'
'''

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.clickhouse.plugins.module_utils.clickhouse import (
    check_clickhouse_driver,
    client_common_argument_spec,
    connect_to_db_via_client,
    execute_query,
    get_main_conn_kwargs,
    get_server_version,
    version_clickhouse_driver,
)


def get_databases(module, client):
    """Get databases.

    Returns a dictionary with database names as keys.
    """
    query = ("SELECT name, engine, data_path, metadata_path, uuid, "
             "engine_full, comment FROM system.databases")
    result = execute_query(module, client, query)

    db_info = {}
    for row in result:
        db_info[row[0]] = {
            "engine": row[1],
            "data_path": row[2],
            "metadata_path": row[3],
            "uuid": str(row[4]),
            "engine_full": row[5],
            "comment": row[6],
        }

    return db_info


def get_clusters(module, client):
    """Get clusters.

    Returns a list with clusters names as top level keys.
    """
    query = ("SELECT cluster, shard_num, shard_weight, replica_num, host_name, "
             "host_address, port, is_local, user, default_database, errors_count, "
             "slowdowns_count, estimated_recovery_time FROM system.clusters")
    result = execute_query(module, client, query)

    cluster_info = {}

    for row in result:
        cluster = row[0]
        shard_num = str(row[1])
        shard_weight = row[2]
        replica_num = str(row[3])
        host_name = row[4]
        host_address = row[5]
        port = row[6]
        is_local = row[7]
        user = row[8]
        default_database = row[9]
        errors_count = row[10]
        slowdowns_count = row[11]
        estimated_recovery_time = row[12]

        # Add cluster if not already there
        if cluster not in cluster_info:
            cluster_info[cluster] = {"shards": {}}

        # Add shard if not already there
        if shard_num not in cluster_info[cluster]["shards"]:
            cluster_info[cluster]["shards"][shard_num] = {
                "shard_weight": shard_weight,
                "replicas": {},
            }

        # Add replica if not already there
        if replica_num not in cluster_info[cluster]["shards"][shard_num]["replicas"]:
            cluster_info[cluster]["shards"][shard_num]["replicas"][replica_num] = {
                "host_name": host_name,
                "host_address": host_address,
                "port": port,
                "is_local": is_local,
                "user": user,
                "default_database": default_database,
                "errors_count": errors_count,
                "slowdowns_count": slowdowns_count,
                "estimated_recovery_time": estimated_recovery_time,
            }

    return cluster_info


def get_roles(module, client):
    """Get roles.

    Returns a dictionary with roles names as keys.
    """
    query = "SELECT name, id, storage FROM system.roles"
    result = execute_query(module, client, query)

    roles_info = {}
    for row in result:
        role_name = row[0]
        roles_info[role_name] = {
            "id": str(row[1]),
            "storage": row[2],
        }
        roles_info[role_name]["grants"] = get_grants(module, client, role_name)

    return roles_info


def get_tables(module, client):
    """Get tables.

    Returns a dictionary with databases name as dictionary,
    and the name of the table in this dictionary is the key.
    """
    query = ("SELECT database, name, uuid, engine, is_temporary, data_paths, "
             "metadata_path, metadata_modification_time, dependencies_database, "
             "dependencies_table, create_table_query, engine_full, as_select, partition_key, "
             "sorting_key, primary_key, sampling_key, storage_policy, total_rows, total_bytes, "
             "parts, active_parts, total_marks, lifetime_rows, lifetime_bytes, comment, "
             "has_own_data, loading_dependencies_database, loading_dependencies_table, "
             "loading_dependent_database, loading_dependent_table FROM system.tables")
    result = execute_query(module, client, query)

    tables_info = {}
    for row in result:
        if row[0] not in tables_info:
            tables_info[row[0]] = {}
        tables_info[row[0]][row[1]] = {
            "uuid": str(row[2]),
            "engine": row[3],
            "is_temporary": row[4],
            "data_paths": row[5],
            "metadata_path": row[6],
            "metadata_modification_time": row[7],
            "dependencies_database": row[8],
            "dependencies_table": row[9],
            "create_table_query": row[10],
            "engine_full": row[11],
            "as_select": row[12],
            "partition_key": row[13],
            "sorting_key": row[14],
            "primary_key": row[15],
            "sampling_key": row[16],
            "storage_policy": row[17],
            "total_rows": row[18],
            "total_bytes": row[19],
            "parts": row[20],
            "active_parts": row[21],
            "total_marks": row[22],
            "lifetime_rows": row[23],
            "lifetime_bytes": row[24],
            "comment": row[25],
            "has_own_data": row[26],
            "loading_dependencies_database": row[27],
            "loading_dependencies_table": row[28],
            "loading_dependent_database": row[29],
            "loading_dependent_table": row[30],
        }

    return tables_info


def get_dictionaries(module, client):
    """Get dictionaries.

    Returns a dictionary with databases name as dictionary,
    and the name of the 'dictionary' in this dictionary is the key.
    """
    query = ("SELECT database, name, uuid, status, origin, key.names, key.types, "
             "attribute.names, attribute.types, bytes_allocated, "
             "hierarchical_index_bytes_allocated, query_count, hit_rate, found_rate, "
             "element_count, load_factor, source, lifetime_min, "
             "lifetime_max, loading_start_time, last_successful_update_time, "
             "loading_duration, last_exception, comment FROM system.dictionaries")
    result = execute_query(module, client, query)

    dictionaries_info = {}
    for row in result:
        dict_database = row[0] if row[0] else 'dict'
        if dict_database not in dictionaries_info:
            dictionaries_info[dict_database] = {}
        dictionaries_info[dict_database][row[1]] = {
            "uuid": str(row[2]),
            "status": row[3],
            "origin": row[4],
            "key.names": row[5],
            "key.types": row[6],
            "attribute.names": row[7],
            "attribute.types": row[8],
            "bytes_allocated": row[9],
            "hierarchical_index_bytes_allocated": row[10],
            "query_count": row[11],
            "hit_rate": row[12],
            "found_rate": row[13],
            "element_count": row[14],
            "load_factor": row[15],
            "source": row[16],
            "lifetime_min": row[17],
            "lifetime_max": row[18],
            "loading_start_time": row[19],
            "last_successful_update_time": row[20],
            "loading_duration": row[21],
            "last_exception": row[22],
        }

    return dictionaries_info


def get_settings(module, client):
    """Get settings.

    Returns a dictionary with settings names as keys.
    """
    query = ("SELECT name, value, changed, description, min, max, readonly, "
             "type, default, alias_for FROM system.settings")
    result = execute_query(module, client, query)

    settings_info = {}
    for row in result:
        settings_info[row[0]] = {
            "value": row[1],
            "changed": row[2],
            "description": row[3],
            "min": row[4],
            "max": row[5],
            "readonly": row[6],
            "type": row[7],
            "default": row[8],
            "alias_for": row[9],
        }

    return settings_info


def get_merge_tree_settings(module, client):
    """Get merge_tree_settings.

    Returns a dictionary with merge_tree_settings names as keys.
    """
    query = ("SELECT name, value, changed, description, min, max, "
             "readonly, type FROM system.merge_tree_settings")
    result = execute_query(module, client, query)

    merge_tree_settings_info = {}
    for row in result:
        merge_tree_settings_info[row[0]] = {
            "value": row[1],
            "changed": row[2],
            "description": row[3],
            "min": row[4],
            "max": row[5],
            "readonly": row[6],
            "type": row[7],
        }

    return merge_tree_settings_info


def get_users(module, client):
    """Get users.

    Returns a dictionary with users names as keys.
    """
    query = ("SELECT name, id, storage, auth_type, auth_params, host_ip, host_names, "
             "host_names_regexp, host_names_like, default_roles_all, "
             "default_roles_list, default_roles_except, grantees_any, "
             "grantees_list, grantees_except, default_database FROM system.users")
    result = execute_query(module, client, query)

    user_info = {}
    for row in result:
        user_name = row[0]
        user_info[user_name] = {
            "id": str(row[1]),
            "storage": row[2],
            "auth_type": row[3],
            "auth_params": row[4],
            "host_ip": row[5],
            "host_names": row[6],
            "host_names_regexp": row[7],
            "host_names_like": row[8],
            "default_roles_all": row[9],
            "default_roles_list": row[10],
            "default_roles_except": row[11],
            "grantees_any": row[12],
            "grantees_list": row[13],
            "grantees_except": row[14],
            "default_database": row[15],

        }

        user_info[user_name]["roles"] = get_user_roles(module, client, user_name)
        user_info[user_name]["grants"] = get_grants(module, client, user_name)

    return user_info


def get_grants(module, client, name):
    """Get grants.

    Return a list of grants.
    """
    query = ("SHOW GRANTS FOR '%s'" % name)
    result = execute_query(module, client, query)
    return [row[0] for row in result]


def get_user_roles(module, client, user_name):
    """Get user roles.

    Returns a list of roles.
    """
    query = ("SELECT granted_role_name FROM system.role_grants "
             "WHERE user_name = '%s'" % user_name)
    result = execute_query(module, client, query)
    return [row[0] for row in result]


def get_settings_profiles(module, client):
    """Get settings profiles.

    Returns a dictionary with profile names as keys.
    """
    query = ("SELECT name, id, storage, num_elements, apply_to_all, apply_to_list, "
             "apply_to_except FROM system.settings_profiles")
    result = execute_query(module, client, query)

    profile_info = {}
    for row in result:
        profile_info[row[0]] = {
            "id": str(row[1]),
            "storage": row[2],
            "num_elements": row[3],
            "apply_to_all": row[4],
            "apply_to_list": row[5],
            "apply_to_except": row[6],
        }

    return profile_info


def get_quotas(module, client):
    """Get quotas.

    Returns a dictionary with quota names as keys.
    """
    query = ("SELECT name, id, storage, keys, durations, apply_to_all, "
             "apply_to_list, apply_to_except FROM system.quotas")
    result = execute_query(module, client, query)

    quota_info = {}
    for row in result:
        quota_info[row[0]] = {
            "id": str(row[1]),
            "storage": row[2],
            "keys": row[3],
            "durations": row[4],
            "apply_to_all": row[5],
            "apply_to_list": row[6],
            "apply_to_except": row[7],
        }

    return quota_info


def get_all_grants(module, client):
    """Get grants.

    Returns a dictionary with users and roles names as keys.
    """
    query = ("SELECT user_name, role_name, access_type, database, "
             "table, column, is_partial_revoke, grant_option FROM system.grants")

    result = execute_query(module, client, query)

    grants_info = {
        'users': {},
        'roles': {},
    }

    for row in result:
        if row[0] is not None:
            dict_name = 'users'
            name = row[0]
            if row[0] not in grants_info[dict_name]:
                grants_info[dict_name][name] = []
        else:
            dict_name = 'roles'
            name = row[1]
            if row[1] not in grants_info[dict_name]:
                grants_info[dict_name][name] = []

        grants_info[dict_name][name].append({
            "access_type": row[2],
            "database": row[3],
            "table": row[4],
            "column": row[5],
            "is_partial_revoke": row[6],
            "grant_option": row[7],
        })

    return grants_info


def get_settings_profile_elements(module, client):
    """Get settings_profile_elements.

    Returns a dictionary with roles, profiles and users names as keys.
    """
    query = ("SELECT profile_name, user_name, role_name, "
             "index, setting_name, value, min, max, writability, "
             "inherit_profile FROM system.settings_profile_elements")
    result = execute_query(module, client, query)

    settings_profile_elements = {'profiles': {},
                                 'users': {},
                                 'roles': {},
                                 }

    for row in result:
        if row[0] is not None:
            dict_name = 'profiles'
            name = row[0]
            if row[0] not in settings_profile_elements[dict_name]:
                settings_profile_elements[dict_name][name] = []
        elif row[1] is not None:
            dict_name = 'users'
            name = row[1]
            if row[1] not in settings_profile_elements[dict_name]:
                settings_profile_elements[dict_name][name] = []
        else:
            dict_name = 'roles'
            name = row[2]
            if row[2] not in settings_profile_elements[dict_name]:
                settings_profile_elements[dict_name][name] = []

        settings_profile_elements[dict_name][name].append({
            "index": row[3],
            "setting_name": row[4],
            "value": row[5],
            "min": row[6],
            "max": row[7],
            "writability": row[8],
            "inherit_profile": row[9],
        })

    return settings_profile_elements


def get_storage_policies(module, client):
    """Get storage_policies.

    Returns a dictionary with storage_policies names as keys.
    """
    query = ("SELECT policy_name, volume_name, volume_priority, "
             "disks, volume_type, max_data_part_size, "
             "move_factor, prefer_not_to_merge FROM system.storage_policies")
    result = execute_query(module, client, query)

    storage_policies_info = {}
    for row in result:
        storage_policies_info[row[0]] = {
            "volume_name": row[1],
            "volume_priority": row[2],
            "disks": row[3],
            "volume_type": row[4],
            "max_data_part_size": row[5],
            "move_factor": row[6],
            "prefer_not_to_merge": row[7],
        }

    return storage_policies_info


def get_driver(module, client):
    """Gets driver information.

    The module and client arguments are here
    to provide a common interface and are ignored.

    Returns its version for now.
    """
    return {"version": version_clickhouse_driver()}


def handle_limit_values(module, supported_ret_vals, limit):
    """Checks if passed limit values match module return values.

    Prints a warning if do not match. Returns a list of stripped vals.
    """
    stripped_vals = []
    for wanted_val in limit:
        wanted_val = wanted_val.strip()

        if wanted_val not in supported_ret_vals:
            msg = ("The passed %s value does not exist in module return values: "
                   "please check the spelling and supported values, and try again" % wanted_val)
            module.warn(msg)
            continue

        stripped_vals.append(wanted_val)

    return stripped_vals


def main():
    # Set up arguments.
    # If there are common arguments shared across several modules,
    # create the common_argument_spec() function under plugins/module_utils/*
    # and invoke here to return a dict with those arguments
    argument_spec = client_common_argument_spec()
    argument_spec.update(
        limit=dict(type='list', elements='str'),
    )

    # Instantiate an object of module class
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    # Assign passed options to variables
    client_kwargs = module.params['client_kwargs']
    # The reason why these arguments are separate from client_kwargs
    # is that we need to protect some sensitive data like passwords passed
    # to the module from logging (see the arguments above with no_log=True);
    # Such data must be passed as module arguments (not nested deep in values).
    main_conn_kwargs = get_main_conn_kwargs(module)

    # Create a mapping between ret values and functions.
    # When adding new values to return, add the value
    # and a corresponding function in this dictionary
    ret_val_func_mapping = {
        'driver': get_driver,
        'version': get_server_version,
        'databases': get_databases,
        'users': get_users,
        'roles': get_roles,
        'settings': get_settings,
        'clusters': get_clusters,
        'dictionaries': get_dictionaries,
        'tables': get_tables,
        'merge_tree_settings': get_merge_tree_settings,
        'quotas': get_quotas,
        'settings_profiles': get_settings_profiles,
        'settings_profile_elements': get_settings_profile_elements,
        'storage_policies': get_storage_policies,
        'grants': get_all_grants,
    }
    # Check if the limit is provided, it contains correct values
    limit = module.params['limit']
    if limit:
        limit = handle_limit_values(module, ret_val_func_mapping.keys(), limit)
    else:
        # If no limit, just gather all ret values
        limit = ret_val_func_mapping.keys()

    # Will fail if no driver informing the user
    check_clickhouse_driver(module)

    # Connect to DB
    client = connect_to_db_via_client(module, main_conn_kwargs, client_kwargs)

    # Get server information
    srv_info = {}
    for item in limit:
        # This will invoke a proper function
        srv_info[item] = ret_val_func_mapping[item](module, client)

    # Close connection
    client.disconnect_connection()

    # Users will get this in JSON output after execution
    module.exit_json(changed=False, **srv_info)


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# Copyright (c) 2019 Zim Kalinowski, (@zikalino)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_galleryimageversion
version_added: "0.1.2"
short_description: Manage Azure SIG Image Version instance
description:
    - Create, update and delete instance of Azure SIG Image Version.
options:
    resource_group:
        description:
            - The name of the resource group.
        required: true
        type: str
    gallery_name:
        description:
            - The name of the Shared Image Gallery in which the Image Definition resides.
        required: true
        type: str
    gallery_image_name:
        description:
            - The name of the gallery Image Definition in which the Image Version is to be created.
        required: true
        type: str
    name:
        description:
            - The name of the gallery Image Version to be created.
            - Needs to follow semantic version name pattern, The allowed characters are digit and period.
            - Digits must be within the range of a 32-bit integer. For example <MajorVersion>.<MinorVersion>.<Patch>.
        required: true
        type: str
    location:
        description:
            - Resource location.
        type: str
    storage_profile:
        description:
            - This is the storage profile of a Gallery Image Version.
            - Required when creating.
        type: dict
        suboptions:
            source_image:
                description:
                    - Reference to managed image or gallery image version
                    - Could be resource ID to managed image, or dictionary containing I(resource_group) and I(name)
                    - Could be resource ID to image version, or dictionary containing I(resource_group),I(gallery_name), I(gallery_image_name) and I(version)
                    - Could be resource ID I(virtual_machine_id) of the source virtual machine.
                    - Mutual exclusive with os_disk and data_disks
                type: raw
            os_disk:
                description:
                    - os disk snapshot
                    - Mutual exclusive with source_image
                type: dict
                suboptions:
                    source:
                        description:
                            - Reference to os disk snapshot.
                            - Could be resource ID.
                            - Could be a dictionary containing I(resource_group) and I(name).
                            - Could be a dictionary containing I(resource_group), I(storage_account), and I(uri)
                              if the snapshot is stored as a PageBlob in a storage account container.
                        type: raw
                    host_caching:
                        description:
                            - host disk caching
                        type: str
                        default: None
                        choices:
                            - None
                            - ReadOnly
                            - ReadWrite
            data_disks:
                description:
                    - list of data disk snapshot
                    - Mutual exclusive with source_image
                type: list
                elements: raw
                suboptions:
                    source:
                        description:
                            - Reference to data disk snapshot. Could be resource ID or dictionary containing I(resource_group) and I(name)
                        type: raw
                    lun:
                        description:
                            - lun of the data disk
                        type: int
                    host_caching:
                        description:
                            - host disk caching
                        type: str
                        default: None
                        choices:
                            - None
                            - ReadOnly
                            - ReadWrite
    publishing_profile:
        description:
            - Publishing profile.
        type: dict
        suboptions:
            target_regions:
                description:
                    - The target regions where the Image Version is going to be replicated to.
                    - This property is updatable.
                type: list
                elements: raw
                suboptions:
                    name:
                        description:
                            - Region name.
                        type: str
                        required: true
                    regional_replica_count:
                        description:
                            - The number of replicas of the Image Version to be created per region.
                            - This property would take effect for a region when regionalReplicaCount is not specified.
                            - This property is updatable.
                        type: int
                    storage_account_type:
                        description:
                            - Storage account type.
                        type: str
                    encryption:
                        description:
                            - Allows users to provide customer managed keys for encrypting the OS and data disks in the gallery artifact.
                        type: dict
                        suboptions:
                            data_disk_images:
                                description:
                                    - A list of encryption specifications for data disk images.
                                type: list
                                elements: dict
                                suboptions:
                                    disk_encryption_set_id:
                                        description:
                                            - A relative URI containing the resource ID of the disk encryption set.
                                        type: str
                                    lun:
                                        description:
                                            - This property specifies the logical unit number of the data disk.
                                            - This value is used to identify data disks within the Virtual Machine and
                                              therefore must be unique for each data disk attached to the Virtual Machine.
                                        type: int
                            os_disk_image:
                                description:
                                    - Contains encryption settings for an OS disk image.
                                type: dict
                                suboptions:
                                    disk_encryption_set_id:
                                        description:
                                            - A relative URI containing the resource ID of the disk encryption set.
                                        type: str
                                    security_profile:
                                        description:
                                            - This property specifies the security profile of an OS disk image.
                                        type: dict
                                        suboptions:
                                            confidential_vm_encryption_type:
                                                description:
                                                    - Confidential VM encryption types.
                                                type: dict
                                                suboptions:
                                                    encrypted_vm_guest_state_only_with_pmk:
                                                        description:
                                                            - VM Guest State Only with PMK.
                                                        type: str
                                                    encrypted_with_cmk:
                                                        description:
                                                            - Encrypted with CMK.
                                                        type: str
                                                    encrypted_with_pmk:
                                                        description:
                                                            - Encrypted with PMK.
                                                        type: str
                                            secure_vm_disk_encryption_set_id:
                                                description:
                                                    - Secure VM disk encryption set id.
                                                type: str
            managed_image:
                description:
                    - Managed image reference, could be resource ID, or dictionary containing I(resource_group) and I(name)
                    - Obsolete since 2.10, use storage_profile instead
                type: raw
            snapshot:
                description:
                    - Source snapshot to be used.
                    - Obsolete since 2.10, use storage_profile instead
                type: raw
            replica_count:
                description:
                    - The number of replicas of the Image Version to be created per region.
                    - This property would take effect for a region when regionalReplicaCount is not specified.
                    - This property is updatable.
                type: int
            exclude_from_latest:
                description:
                    If I(exclude_from_latest=true), Virtual Machines deployed from the latest version of the Image Definition won't use this Image Version.
                type: bool
            end_of_life_date:
                description:
                    - The end of life date of the gallery Image Version.
                    - This property can be used for decommissioning purposes.
                    - This property is updatable. Format should be according to ISO-8601, for instance "2019-06-26".
                type: str
            storage_account_type:
                description:
                    - Specifies the storage account type to be used to store the image.
                    - This property is not updatable.
                type: str
                choices:
                    - Standard_LRS
                    - Standard_ZRS
    timeout:
        description:
            - Set the timeout (minute) period for creating an Image Version.
        type: int
        default: 10
    state:
        description:
            - Assert the state of the GalleryImageVersion.
            - Use C(present) to create or update an GalleryImageVersion and C(absent) to delete it.
        default: present
        choices:
            - absent
            - present
        type: str
extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags
author:
    - Zim Kalinowski (@zikalino)

'''

EXAMPLES = '''
- name: Create a gallery image version form a managed image
  azure_rm_galleryimageversion:
    resource_group: myResourceGroup
    gallery_name: myGallery
    gallery_image_name: myGalleryImage
    name: 1.1.0
    location: East US
    publishing_profile:
      end_of_life_date: "2020-10-01t00:00:00+00:00"
      exclude_from_latest: true
      replica_count: 4
      storage_account_type: Standard_LRS
      target_regions:
        - name: West US
          regional_replica_count: 1
        - name: East US
          regional_replica_count: 3
          storage_account_type: Standard_LRS
    storage_profile:
      source_image: /subscriptions/sub123/resourceGroups/group123/providers/Microsoft.Compute/images/myOsImage

- name: Create a gallery image version from another gallery image version
  azure_rm_galleryimageversion:
    resource_group: myResourceGroup
    gallery_name: myGallery
    gallery_image_name: myGalleryImage
    name: 1.2.0
    location: East US
    publishing_profile:
      end_of_life_date: "2020-10-01t00:00:00+00:00"
      exclude_from_latest: true
      replica_count: 4
      storage_account_type: Standard_LRS
      target_regions:
        - name: West US
          regional_replica_count: 1
        - name: East US
          regional_replica_count: 3
          storage_account_type: Standard_LRS
    storage_profile:
      source_image:
        version: 1.1.0
        gallery_name: myGallery2
        gallery_image_name: myGalleryImage2

- name: Create gallery image by using one os dist snapshot and zero or many data disk snapshots
  azure_rm_galleryimageversion:
    resource_group: myRsourceGroup
    gallery_name: myGallery
    gallery_image_name: myGalleryImage
    name: 3.4.0
    location: East  US
    publishing_profile:
      end_of_life_date: "2020-10-01t00:00:00+00:00"
      exclude_from_latest: true
      replica_count: 1
      storage_account_type: Standard_LRS
      target_regions:
        - name: East US
          regional_replica_count: 1
          storage_account_type: Standard_LRS
    storage_profile:
      os_disk:
        source: "/subscriptions/mySub/resourceGroups/myGroup/providers/Microsoft.Compute/snapshots/os_snapshot_vma"
      data_disks:
        - lun: 0
          source:
            name: data_snapshot_vma
        - lun: 1
          source: "/subscriptions/mySub/resourceGroups/myGroup/providers/Microsoft.Compute/snapshots/data_snapshot_vmb"

- name: Create gallery image by using a os disk snapshot stored in Storage Account container
  azure_rm_galleryimageversion:
    resource_group: myResourceGroup
    gallery_name: myGallery
    gallery_image_name: myGalleryImage
    name: 3.4.0
    location: East  US
    publishing_profile:
      end_of_life_date: "2020-10-01t00:00:00+00:00"
      exclude_from_latest: true
      replica_count: 1
      storage_account_type: Standard_LRS
      target_regions:
        - name: East US
          regional_replica_count: 1
          storage_account_type: Standard_LRS
    storage_profile:
      os_disk:
        source:
          resource_group: myResourceGroup
          storage_account: myStorageAccount
          uri: "https://myStorageAccount.blob.core.windows.net/myContainer/myImage.vhd"
'''

RETURN = '''
id:
    description:
        - Resource ID.
    returned: always
    type: str
    sample: "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/myResourceGroup/providers/Microsoft.Compute/galleries/myGalle
           ry1283/images/myImage/versions/10.1.3"
'''

try:
    import time
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ext import AzureRMModuleBaseExt
    from azure.core.polling import LROPoller
    from azure.core.exceptions import ResourceNotFoundError
except ImportError:
    pass


class Actions:
    NoAction, Create, Update, Delete = range(4)


class AzureRMGalleryImageVersions(AzureRMModuleBaseExt):
    def __init__(self):
        self.module_arg_spec = dict(
            resource_group=dict(
                type='str',
                required=True
            ),
            gallery_name=dict(
                type='str',
                required=True
            ),
            gallery_image_name=dict(
                type='str',
                required=True
            ),
            name=dict(
                type='str',
                required=True
            ),
            tags=dict(
                type='dict',
            ),
            location=dict(
                type='str',
            ),
            storage_profile=dict(
                type='dict',
                mutually_exclusive=[('source_image', 'os_disk', 'data_disks')],
                options=dict(
                    source_image=dict(
                        type='raw',
                    ),
                    os_disk=dict(
                        type='dict',
                        options=dict(
                            source=dict(
                                type='raw',
                            ),
                            host_caching=dict(
                                type='str',
                                default="None",
                                choices=["ReadOnly", "ReadWrite", "None"]
                            )
                        )
                    ),
                    data_disks=dict(
                        type='list',
                        elements='raw',
                        options=dict(
                            lun=dict(
                                type='int'
                            ),
                            source=dict(
                                type='raw',
                            ),
                            host_caching=dict(
                                type='str',
                                default="None",
                                choices=["ReadOnly", "ReadWrite", "None"]
                            )
                        )
                    )
                )
            ),
            publishing_profile=dict(
                type='dict',
                options=dict(
                    target_regions=dict(
                        type='list',
                        elements='raw',
                        options=dict(
                            name=dict(
                                type='str',
                                required=True,
                            ),
                            regional_replica_count=dict(
                                type='int',
                            ),
                            storage_account_type=dict(
                                type='str',
                            ),
                            encryption=dict(
                                type='dict',
                                options=dict(
                                    data_disk_images=dict(
                                        type='list',
                                        elements='dict',
                                        options=dict(
                                            disk_encryption_set_id=dict(
                                                type='str',
                                            ),
                                            lun=dict(
                                                type='int'
                                            )
                                        )
                                    ),
                                    os_disk_image=dict(
                                        type='dict',
                                        options=dict(
                                            disk_encryption_set_id=dict(
                                                type='str',
                                            ),
                                            security_profile=dict(
                                                type='dict',
                                                options=dict(
                                                    confidential_vm_encryption_type=dict(
                                                        type='dict',
                                                        options=dict(
                                                            encrypted_vm_guest_state_only_with_pmk=dict(
                                                                type='str',
                                                            ),
                                                            encrypted_with_cmk=dict(
                                                                type='str',
                                                            ),
                                                            encrypted_with_pmk=dict(
                                                                type='str',
                                                            )
                                                        )
                                                    ),
                                                    secure_vm_disk_encryption_set_id=dict(
                                                        type='str',
                                                    )
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                        )
                    ),
                    managed_image=dict(
                        type='raw',
                    ),
                    snapshot=dict(
                        type='raw',
                    ),
                    replica_count=dict(
                        type='int',
                    ),
                    exclude_from_latest=dict(
                        type='bool',
                    ),
                    end_of_life_date=dict(
                        type='str',
                    ),
                    storage_account_type=dict(
                        type='str',
                        choices=['Standard_LRS',
                                 'Standard_ZRS']
                    )
                )
            ),
            timeout=dict(
                type='int',
                default=10
            ),
            state=dict(
                type='str',
                default='present',
                choices=['present', 'absent']
            )
        )

        self.resource_group = None
        self.gallery_name = None
        self.gallery_image_name = None
        self.name = None
        self.gallery_image_version = None
        self.tags = None
        self.state = None
        self.timeout = None

        self.body = dict()
        self.results = dict(changed=False)
        self.to_do = Actions.NoAction

        super(AzureRMGalleryImageVersions, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                          supports_check_mode=True,
                                                          supports_tags=True)

    def exec_module(self, **kwargs):
        for key in list(self.module_arg_spec.keys()):
            if hasattr(self, key):
                setattr(self, key, kwargs[key])
                if key == 'tags':
                    self.body[key] = kwargs[key]
            elif kwargs[key] is not None:
                if key == 'location':
                    self.body['location'] = kwargs[key]
                elif key == 'storage_profile':
                    self.body['storage_profile'] = {}
                    if kwargs[key].get('source_image') is not None:
                        self.body['storage_profile']['source'] = {}
                        if isinstance(kwargs[key].get('source_image'), str):
                            self.body['storage_profile']['source']['id'] = kwargs[key].get('source_image')
                        elif isinstance(kwargs[key].get('source_image'), dict):
                            if kwargs[key]['source_image'].get('id') is not None:
                                self.body['storage_profile']['source']['id'] = kwargs[key]['source_image'].get('id')
                            elif kwargs[key]['source_image'].get('virtual_machine_id') is not None:
                                self.body['storage_profile']['source']['virtualMachineId'] = kwargs[key]['source_image'].get('virtual_machine_id')
                            elif kwargs[key]['source_image'].get('resource_group') is not None and kwargs[key]['source_image'].get('name') is not None:
                                self.body['storage_profile']['source']['id'] = ('/subscriptions/' +
                                                                                self.subscription_id +
                                                                                '/resourceGroups/' +
                                                                                kwargs[key]['source_image'].get('resource_group') +
                                                                                '/providers/Microsoft.Compute/images/' +
                                                                                kwargs[key]['source_image'].get('name'))
                            elif (kwargs[key]['source_image'].get('resource_group') is not None and
                                  kwargs[key]['source_image'].get('gallery_name') is not None and
                                  kwargs[key]['source_image'].get('gallery_image_name') is not None and kwargs[key]['source_image'].get('version') is not None):
                                self.body['storage_profile']['source']['id'] = ('/subscriptions/' +
                                                                                self.subscription_id +
                                                                                '/resourceGroups/' +
                                                                                kwargs[key]['source_image'].get('resource_group') +
                                                                                '/providers/Microsoft.Compute/galleries/' +
                                                                                kwargs[key]['source_image'].get('gallery_name') +
                                                                                '/images/' +
                                                                                kwargs[key]['source_image'].get('gallery_image_name') +
                                                                                '/versions/' +
                                                                                kwargs[key]['source_image'].get('version'))
                            else:
                                self.fail("The source_image parameters config errors")
                        else:
                            self.fail("The source_image parameters config errors")
                    if kwargs[key].get('os_disk') is not None:
                        self.body['storage_profile']['os_disk_image'] = {}
                        if kwargs[key]['os_disk'].get('host_caching') is not None:
                            self.body['storage_profile']['os_disk_image']['hostCaching'] = kwargs[key]['os_disk'].get('host_caching')
                        if kwargs[key]['os_disk'].get('source') is not None:
                            self.body['storage_profile']['os_disk_image']['source'] = {}
                            if isinstance(kwargs[key]['os_disk']['source'], str):
                                self.body['storage_profile']['os_disk_image']['source']['id'] = kwargs[key]['os_disk']['source']
                            elif isinstance(kwargs[key]['os_disk']['source'], dict):
                                if kwargs[key]['os_disk']['source'].get('id') is not None:
                                    self.body['storage_profile']['os_disk_image']['source']['id'] = kwargs[key]['os_disk']['source'].get('id')
                                elif kwargs[key]['os_disk']['source'].get('resource_group') is not None and \
                                        kwargs[key]['os_disk']['source'].get('name') is not None:
                                    resource_group = kwargs[key]['os_disk']['source'].get('resource_group')
                                    self.body['storage_profile']['os_disk_image']['source']['id'] = ('/subscriptions/' +
                                                                                                     self.subscription_id +
                                                                                                     '/resourceGroups/' +
                                                                                                     resource_group +
                                                                                                     '/providers/Microsoft.Compute/snapshots/' +
                                                                                                     kwargs[key]['os_disk']['source'].get('name'))
                                elif kwargs[key]['os_disk']['source'].get('uri') is not None and \
                                        kwargs[key]['os_disk']['source'].get('resource_group') is not None and \
                                        kwargs[key]['os_disk']['source'].get('storage_account') is not None:
                                    resource_group = kwargs[key]['os_disk']['source'].get('resource_group')
                                    storage_account = kwargs[key]['os_disk']['source'].get('storage_account')
                                    self.body['storage_profile']['os_disk_image']['source']['storageAccountId'] = ('/subscriptions/' +
                                                                                                                   self.subscription_id +
                                                                                                                   '/resourceGroups/' +
                                                                                                                   resource_group +
                                                                                                                   '/providers/Microsoft.Storage' +
                                                                                                                   '/storageAccounts/' +
                                                                                                                   storage_account)
                                    self.body['storage_profile']['os_disk_image']['source']['uri'] = kwargs[key]['os_disk']['source'].get('uri')
                                else:
                                    self.fail("The os_disk.source parameters config errors")

                            else:
                                self.fail("The os_disk.source parameters config errors")

                    if kwargs[key].get('data_disks') is not None:
                        self.body['storage_profile']['data_disk_images'] = []
                        data_disk = {}
                        for item in kwargs[key].get('data_disks'):
                            if item.get('lun') is not None:
                                data_disk['lun'] = item['lun']
                            if item.get('source') is not None:
                                data_disk['source'] = {}
                                if isinstance(item.get('source'), str):
                                    data_disk['source']['id'] = item.get('source')
                                elif isinstance(item.get('source'), dict):
                                    if item['source'].get('id') is not None:
                                        data_disk['source']['id'] = item['source'].get('id')
                                    elif item['source'].get('resource_group') is not None and item['source'].get('name') is not None:
                                        data_disk['source']['id'] = ('/subscriptions/' +
                                                                     self.subscription_id +
                                                                     '/resourceGroups/' +
                                                                     item['source'].get('resource_group') +
                                                                     '/providers/Microsoft.Compute/snapshots/' +
                                                                     item['source'].get('name'))
                                    else:
                                        self.fail("The data_disk.source parameters config errors")
                                else:
                                    self.fail("The data_disk.source parameters config errors")

                            if item.get('host_caching') is not None:
                                data_disk['hostCaching'] = item['host_caching']

                            self.body['storage_profile']['data_disk_images'].append(data_disk)

                elif key == 'publishing_profile':
                    self.body['publishing_profile'] = {}
                    if kwargs['publishing_profile'].get('target_regions') is not None:
                        self.body['publishing_profile']['target_regions'] = []
                        for item in kwargs['publishing_profile']['target_regions']:
                            target_regions = {}
                            for value in item.keys():
                                if value == 'name':
                                    target_regions[value] = item[value]
                                elif value == 'regional_replica_count':
                                    target_regions['regional_replica_count'] = item[value]
                                elif value == 'storage_account_type':
                                    target_regions['storage_account_type'] = item[value]
                                elif value == 'encryption':
                                    target_regions['encryption'] = {}
                                    if item[value].get('data_disk_images') is not None:
                                        target_regions['encryption']['data_disk_images'] = []
                                        for tt in item[value]['data_disk_images']:
                                            disk_image = {}
                                            if tt.get('lun') is not None:
                                                disk_image['lun'] = tt['lun']
                                            if tt.get('disk_encryption_set_id') is not None:
                                                disk_image['disk_encryption_set_id'] = tt['disk_encryption_set_id']
                                            target_regions['encryption']['data_disk_images'].append(disk_image)

                                    if item['encryption'].get('os_disk_image') is not None:
                                        target_regions['encryption']['os_disk_image'] = {}
                                        if item['encryption']['os_disk_image'].get('disk_encryption_set_id') is not None:
                                            disk_encryption_set_id = item['encryption']['os_disk_image']['disk_encryption_set_id']
                                            target_regions['encryption']['os_disk_image']['disk_encryption_set_id'] = disk_encryption_set_id
                                        if item['encryption']['os_disk_image'].get('security_profile') is not None:
                                            target_regions['encryption']['os_disk_image']['security_profile'] = {}
                                            if item['encryption']['os_disk_image']['security_profile'].get('secure_vm_disk_encryption_set_id') is not None:
                                                secure_id = item['encryption']['os_disk_image']['security_profile']['secure_vm_disk_encryption_set_id']
                                                target_regions['encryption']['os_disk_image']['security_profile']['secure_vm_dis_encryption_set_id'] = secure_id

                                            if item['encryption']['os_disk_image']['security_profile'].get('confidential_vm_encryption_type') is not None:
                                                target_regions['encryption']['os_disk_image']['security_profile']['confidential_vm_encryption_type'] = {}
                                                security = item['encryption']['os_disk_image']['security_profile']['confidential_vm_encryption_type']
                                                tt = target_regions['encryption']['os_disk_image']['security_profile']['confidential_vm_encryption_type']
                                                if security.get('encrypted_vm_guest_state_only_with_pmk') is not None:
                                                    tt['encrypted_vm_guestState_only_with_pmk'] = security.get('encrypted_vm_guest_state_only_with_pmk')
                                                if security.get('encrypted_with_cmk') is not None:
                                                    tt['encrypted_with_cmk'] = security.get('encrypted_with_cmk')
                                                if security.get('encrypted_with_pmk') is not None:
                                                    tt['encrypted_with_pmk'] = security.get('encrypted_with_pmk')
                            self.body['publishing_profile']['target_regions'].append(target_regions)
                    if kwargs[key].get('managed_image') is not None:
                        if isinstance(kwargs[key]['managed_image'], str):
                            self.body['publishing_profile']['managed_image'] = kwargs[key]['managed_image']
                        elif isinstance(kwargs[key]['managed_image'], dict):
                            if kwargs[key]['managed_image'].get('id') is not None:
                                self.body['publishing_profile']['managed_image'] = kwargs[key]['managed_image']['id']
                            elif kwargs[key]['managed_image'].get('resource_group') is not None and kwargs[key]['managed_image'].get('name') is not None:
                                self.body['publishing_profile']['managed_image'] = ('/subscriptions/' +
                                                                                    self.subscription_id +
                                                                                    '/resourceGroups/' +
                                                                                    kwargs[key]['managed_image'].get('resource_group') +
                                                                                    '/providers/Microsoft.Compute/images/' +
                                                                                    kwargs[key]['managed_image'].get('name'))
                            else:
                                self.fail("The managed_image parameters config errors")
                        else:
                            self.fail("The managed_image parameters config errors")
                    if kwargs[key].get('snapshot') is not None:
                        if isinstance(kwargs[key].get('snapshot'), str):
                            self.body['publishing_profile']['snapshot'] = kwargs[key].get('snapshot')
                        elif isinstance(kwargs[key].get('snapshot'), dict):
                            if kwargs[key]['snapshot'].get('id') is not None:
                                self.body['publishing_profile']['snapshot'] = kwargs[key]['snapshot'].get('id')
                            elif kwargs[key]['snapshot'].get('resource_group') is not None and kwargs[key]['snapshot'].get('name') is not None:
                                self.body['publishing_profile']['snapshot'] = ('/subscriptions/' +
                                                                               self.subscription_id +
                                                                               '/resourceGroups/' +
                                                                               kwargs[key]['snapshot'].get('resource_group') +
                                                                               '/providers/Microsoft.Compute/snapshots/' +
                                                                               kwargs[key]['snapshot'].get('name'))
                            else:
                                self.fail("The managed_image parameters config errors")
                        else:
                            self.fail("The managed_image parameters config errors")
                    if kwargs[key].get('replica_count') is not None:
                        self.body['publishing_profile']['replica_count'] = kwargs[key].get('replica_count')
                    if kwargs[key].get('exclude_from_latest') is not None:
                        self.body['publishing_profile']['exclude_from_latest'] = kwargs[key].get('exclude_from_latest')
                    if kwargs[key].get('end_of_life_date') is not None:
                        self.body['publishing_profile']['end_of_life_date'] = kwargs[key].get('end_of_life_date')
                    if kwargs[key].get('storage_account_type') is not None:
                        self.body['publishing_profile']['storage_account_type'] = kwargs[key].get('storage_account_type')

        # keep backward compatibility
        snapshot = self.body.get('publishing_profile', {}).pop('snapshot', None)
        if snapshot is not None:
            self.body.setdefault('storage_profile', {}).setdefault('os_disk_image', {}).setdefault('source', {})['id'] = snapshot
        managed_image = self.body.get('publishing_profile', {}).pop('managed_image', None)
        if managed_image:
            self.body.setdefault('storage_profile', {}).setdefault('source', {})['id'] = managed_image

        old_response = None
        response = None

        resource_group = self.get_resource_group(self.resource_group)

        if 'location' not in self.body:
            self.body['location'] = resource_group.location

        old_response = self.get_resource()

        if not old_response:
            self.log("GalleryImageVersion instance doesn't exist")

            if self.state == 'absent':
                self.log("Old instance didn't exist")
            else:
                self.to_do = Actions.Create
        else:
            self.log('GalleryImageVersion instance already exists')

            if self.state == 'absent':
                self.to_do = Actions.Delete
            else:
                update_tags, newtags = self.update_tags(old_response.get('tags', dict()))
                if update_tags:
                    self.tags = newtags
                    self.body['tags'] = self.tags
                    self.to_do = Actions.Update
                if self.body.get('publishing_profile') is not None:
                    for key in self.body['publishing_profile'].keys():
                        if key == 'target_regions':
                            result = dict(compare=[])
                            modifies = {'/*/name': {'updatable': True, 'comparison': 'location'}}
                            if not self.default_compare(modifies, self.body['publishing_profile'][key], old_response['publishing_profile'][key], '', result):
                                self.to_do = Actions.Update
                        elif key == 'end_of_life_date':
                            if self.body['publishing_profile'][key][:10].lower() != old_response['publishing_profile'].get(key)[:10].lower():
                                self.to_do = Actions.Update
                        elif self.body['publishing_profile'].get(key) != old_response['publishing_profile'].get(key):
                            self.to_do = Actions.Update

        if (self.to_do == Actions.Create) or (self.to_do == Actions.Update):
            self.log('Need to Create / Update the GalleryImageVersion instance')

            if self.check_mode:
                self.results['changed'] = True
                return self.results

            response = self.create_update_resource()

            self.results['changed'] = True
            self.log('Creation / Update done')
        elif self.to_do == Actions.Delete:
            self.log('GalleryImageVersion instance deleted')
            self.results['changed'] = True

            if self.check_mode:
                return self.results

            self.delete_resource()
        else:
            self.log('GalleryImageVersion instance unchanged')
            self.results['changed'] = False
            response = old_response

        if response:
            self.results["id"] = response["id"]

        return self.results

    def create_update_resource(self):
        # self.log('Creating / Updating the GalleryImageVersion instance {0}'.format(self.))

        response = None
        try:
            if self.to_do == Actions.Create:
                response = self.image_version_client.gallery_image_versions.begin_create_or_update(self.resource_group, self.gallery_name,
                                                                                                   self.gallery_image_name, self.name, self.body)
            else:
                response = self.image_version_client.gallery_image_versions.begin_update(self.resource_group,
                                                                                         self.gallery_name, self.gallery_image_name, self.name, self.body)

        except Exception as exc:
            self.log('Error attempting to create the GalleryImageVersion instance.')
            self.fail('Error creating the GalleryImageVersion instance: {0}'.format(str(exc)))

        if isinstance(response, LROPoller):
            response = self.get_poller_result(response).as_dict()

        i = 0
        while response['provisioning_state'] == 'Creating':
            time.sleep(60)
            response = self.get_resource()
            i = i + 1
            if i == self.timeout:
                self.fail("Create or Updating encountered an exception, wait 10 minutes when the status is still 'creating'")

        return response

    def delete_resource(self):
        # self.log('Deleting the GalleryImageVersion instance {0}'.format(self.))
        try:
            response = self.image_version_client.gallery_image_versions.begin_delete(self.resource_group, self.gallery_name, self.gallery_image_name, self.name)
        except Exception as e:
            self.log('Error attempting to delete the GalleryImageVersion instance.')
            self.fail('Error deleting the GalleryImageVersion instance: {0}'.format(str(e)))
        return True

    def get_resource(self):
        # self.log('Checking if the GalleryImageVersion instance {0} is present'.format(self.))
        response = None
        try:
            response = self.image_version_client.gallery_image_versions.get(self.resource_group, self.gallery_name, self.gallery_image_name, self.name)
        except ResourceNotFoundError:
            self.log('Could not get the gallery image verison')
        return self.format_item(response)

    def format_item(self, item):
        if item is None:
            return None
        else:
            return item.as_dict()


def main():
    AzureRMGalleryImageVersions()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Joshua Coronado (@joshuajcoronado) <joshua@coronado.io>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {
    "metadata_version": "1.1",
    "status": ["deprecated"],
    "supported_by": "community",
}

DOCUMENTATION = r"""
author:
  - Joshua Coronado (@joshuajcoronado)
deprecated:
  alternative: cisco.meraki.networks_webhooks_payload_templates
  removed_in: 3.0.0
  why: Updated modules released with increased functionality
description:
  - Allows for querying, deleting, creating, and updating of webhook payload templates.
extends_documentation_fragment: cisco.meraki.meraki
module: meraki_webhook_payload_template
options:
  body:
    description:
      - The liquid template used for the body of the webhook message.
    type: str
  headers:
    default: []
    description:
      - List of the liquid templates used with the webhook headers.
    elements: dict
    suboptions:
      name:
        description:
          - The name of the header template.
        type: str
      template:
        description:
          - The liquid template for the headers
        type: str
    type: list
  name:
    description:
      - Name of the template.
    type: str
  net_id:
    description:
      - ID of network containing access points.
    type: str
  net_name:
    description:
      - Name of network containing access points.
    type: str
  state:
    choices:
      - absent
      - query
      - present
    default: query
    description:
      - Specifies whether payload template should be queried, created, modified, or
        deleted.
    type: str
short_description: Manage webhook payload templates for a network in the Meraki cloud
"""

EXAMPLES = r"""
- name: Query all configuration templates
  meraki_webhook_payload_template:
    auth_key: abc12345
    org_name: YourOrg
    state: query
  delegate_to: localhost
- name: Query specific configuration templates
  meraki_webhook_payload_template:
    auth_key: abc12345
    org_name: YourOrg
    state: query
    name: Twitter
  delegate_to: localhost
- name: Create payload template
  meraki_webhook_payload_template:
    auth_key: abc12345
    org_name: YourOrg
    state: query
    name: TestTemplate
    body: Testbody
    headers:
      - name: testheader
        template: testheadertemplate
  delegate_to: localhost
- name: Delete a configuration template
  meraki_config_template:
    auth_key: abc123
    state: absent
    org_name: YourOrg
    name: TestTemplate
  delegate_to: localhost
"""

RETURN = r"""
data:
    description: Information about queried object.
    returned: success
    type: complex
    contains:
        name:
            description:
                - The name of the template
            returned: success
            type: str
            sample: testTemplate
        body:
            description:
                - The liquid template used for the body of the webhook message.
            returned: success
            type: str
            sample: {"event_type":"{{ alertTypeId }}","client_payload":{"text":"{{ alertData }}"}}
        headers:
            description: List of the liquid templates used with the webhook headers.
            returned: success
            type: list
            contains:
                name:
                    description:
                        - The name of the template
                    returned: success
                    type: str
                    sample: testTemplate
                template:
                    description:
                        - The liquid template for the header
                    returned: success
                    type: str
                    sample: "Bearer {{ sharedSecret }}"
"""

from ansible.module_utils.basic import AnsibleModule, json
from ansible_collections.cisco.meraki.plugins.module_utils.network.meraki.meraki import (
    MerakiModule,
    meraki_argument_spec,
)


def get_webhook_payload_templates(meraki, net_id):
    path = meraki.construct_path("get_all", net_id=net_id)
    response = meraki.request(path, "GET")
    if meraki.status != 200:
        meraki.fail_json(msg="Unable to get webhook payload templates")
    return response


def delete_template(meraki, net_id, template_id):
    changed = True
    if meraki.check_mode:
        return {}, changed
    else:
        path = meraki.construct_path(
            "update", net_id=net_id, custom={"template_id": template_id}
        )
        response = meraki.request(path, method="DELETE")
        if meraki.status != 204:
            meraki.fail_json(msg="Unable to remove webhook payload templates")
        return response, changed


def create_template(meraki, net_id, template):
    changed = True

    if meraki.check_mode:
        return template, changed
    else:
        path = meraki.construct_path("get_all", net_id=net_id)
        response = meraki.request(path, "POST", payload=json.dumps(template))
        if meraki.status != 201:
            meraki.fail_json(msg="Unable to create webhook payload template")
        return response, changed


def update_template(meraki, net_id, template, payload):
    changed = False

    if template["body"] != payload["body"]:
        changed = True

    if meraki.is_update_required(template["headers"], payload["headers"]):
        changed = True

    if changed:
        meraki.generate_diff(template, payload)
        if meraki.check_mode:
            return payload, changed
        else:
            path = meraki.construct_path(
                "update",
                net_id=net_id,
                custom={"template_id": template["payloadTemplateId"]},
            )
            response = meraki.request(
                path, method="PUT", payload=json.dumps(payload)
            )
            if meraki.status != 200:
                meraki.fail_json(
                    msg="Unable to update webhook payload template"
                )
            return response, changed

    return template, changed


def main():

    # define the available arguments/parameters that a user can pass to
    # the module
    argument_spec = meraki_argument_spec()
    argument_spec.update(
        state=dict(
            type="str", choices=["absent", "query", "present"], default="query"
        ),
        name=dict(type="str", default=None),
        net_name=dict(type="str"),
        net_id=dict(type="str"),
        body=dict(type="str", default=None),
        headers=dict(
            type="list",
            default=[],
            elements="dict",
            options=dict(
                name=dict(type="str"),
                template=dict(type="str"),
            ),
        ),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )
    meraki = MerakiModule(module, function="webhook_payload_template")
    meraki.params["follow_redirects"] = "all"

    query_all_urls = {
        "webhook_payload_template": "/networks/{net_id}/webhooks/payloadTemplates"
    }
    update_urls = {
        "webhook_payload_template": "/networks/{net_id}/webhooks/payloadTemplates/{template_id}"
    }

    meraki.url_catalog["get_all"].update(query_all_urls)
    meraki.url_catalog["update"] = update_urls

    org_id = meraki.params["org_id"]
    if meraki.params["org_name"]:
        org_id = meraki.get_org_id(meraki.params["org_name"])
    net_id = meraki.params["net_id"]

    if net_id is None:
        if meraki.params["net_name"] is not None:
            nets = meraki.get_nets(org_id=org_id)
            net_id = meraki.get_net_id(
                net_name=meraki.params["net_name"], data=nets
            )

    templates = {
        template["name"]: template
        for template in get_webhook_payload_templates(meraki, net_id)
    }

    if meraki.params["state"] == "query":
        meraki.result["changed"] = False

        if meraki.params["name"]:
            if meraki.params["name"] in templates:
                meraki.result["data"] = templates[meraki.params["name"]]
            else:
                meraki.fail_json(
                    msg="Unable to get webhook payload template named: {0}".format(
                        meraki.params["name"]
                    )
                )
        else:
            meraki.result["data"] = templates

    elif meraki.params["state"] == "present":
        if meraki.params["name"] is None:
            meraki.fail_json(msg="name is a required parameter")

        if meraki.params["body"] is None:
            meraki.fail_json(
                msg="body is a required parameter when state is present"
            )

        headers = []

        for header in meraki.params["headers"]:
            for key in ["name", "template"]:
                if key not in header:
                    meraki.fail_json(
                        msg="{0} is a required parameter for a header".format(
                            key
                        )
                    )
                if not header[key]:
                    meraki.fail_json(
                        msg="{0} in header must be a string".format(key)
                    )
            headers.append(
                dict(name=header["name"], template=header["template"])
            )

        payload = {
            "name": meraki.params["name"],
            "body": meraki.params["body"],
            "headers": meraki.params["headers"],
        }

        if meraki.params["name"] in templates:
            (
                meraki.result["data"],
                meraki.result["changed"],
            ) = update_template(
                meraki, net_id, templates[meraki.params["name"]], payload
            )
        else:
            (
                meraki.result["data"],
                meraki.result["changed"],
            ) = create_template(meraki, net_id, payload)

    elif meraki.params["state"] == "absent":
        if meraki.params["name"] in templates:
            (
                meraki.result["data"],
                meraki.result["changed"],
            ) = delete_template(
                meraki,
                net_id,
                templates[meraki.params["name"]]["payloadTemplateId"],
            )
        else:
            meraki.result["changed"] = False
            meraki.result["data"] = {}

    # in the event of a successful module execution, you will want to
    # simple AnsibleModule.exit_json(), passing the key/value results
    meraki.exit_json(**meraki.result)


if __name__ == "__main__":
    main()
